import contextlib
import fnmatch
import glob
import json
import os
import re
import sys
from datetime import datetime, timezone
from functools import reduce
from typing import Any, Callable, Dict, List, Optional, Set, Tuple
from uuid import UUID

import click
import yaml
from box import Box
from marshmallow import ValidationError
from pycarlo.core import Client, Query
from tabulate import tabulate
from yaml import SafeLoader

from montecarlodata.config import Config
from montecarlodata.errors import complain_and_abort, manage_errors
from montecarlodata.fs_utils import mkdirs
from montecarlodata.iac.client import MonteCarloConfigTemplateClient
from montecarlodata.iac.schemas import (
    ConfigTemplateUpdateState,
    ProjectConfig,
    ResourceModification,
)
from montecarlodata.iac.utils import (
    has_montecarlo_property,
    has_namespace_property,
    is_dbt_schema,
)
from montecarlodata.settings import (
    DBT_CONFIG_YAML,
    DBT_CONFIG_YML,
    PROJECT_CONFIG_FILENAME_YAML,
    PROJECT_CONFIG_FILENAME_YML,
    TARGET_DIRECTORY_NAME,
)
from montecarlodata.utils import GqlWrapper


class TemplateLoader(SafeLoader):
    def construct_mapping(self, node, deep=False):
        # check if there are duplicate keys first
        keys = set()
        for key_node, value_node in node.value:
            if key_node.tag == "tag:yaml.org,2002:merge":
                # skip merge nodes
                # we don't want to detect duplicates for them either since they can
                # be used to declare defaults that are later overridden, so we  avoid flattening
                # them before this check
                continue
            key = self.construct_object(key_node, deep=deep)
            if key in keys:
                raise ValueError(f'Duplicate "{key}" key found in YAML.')
            keys.add(key)

        mapping = super(TemplateLoader, self).construct_mapping(node, deep=deep)
        # Annotate each node with line number and filename
        # Add 1 so line numbering starts at 1
        mapping["yaml_line"] = node.start_mark.line + 1
        mapping["yaml_file_name"] = node.start_mark.name
        return mapping


class MonteCarloConfigService:
    NS_HEADERS = ("Namespace", "Last Update Time")
    MORE_NS_MESSAGE = "There are more namespaces available. Increase the limit to view them."
    DBT_REF_PATTERN = re.compile(r"\[\[\s*ref\(['\"](\w+)['\"]\)\s*\]\]")
    SQL_KEYS = {"sql", "sampling_sql", "variables"}

    def __init__(
        self,
        config: Config,
        pycarlo_client: Client,
        command_name: str,
        request_wrapper: Optional[GqlWrapper] = None,
        project_dir: Optional[str] = None,
        dbt_manifest_path: Optional[str] = None,
        print_func: Optional[Callable] = None,
    ):
        self._command_name = command_name
        request_wrapper = request_wrapper or GqlWrapper(config, command_name=self._command_name)
        self.client = MonteCarloConfigTemplateClient(request_wrapper)
        self._pycarlo_client = pycarlo_client
        self._print_func = print_func or click.echo
        self.project_dir = os.path.abspath(project_dir) if project_dir else os.getcwd()
        self.dbt_manifest_path = os.path.abspath(dbt_manifest_path) if dbt_manifest_path else None
        self._abort_on_error = True

        # Nicer formatting for multiline strings
        yaml.add_representer(str, self._str_presenter)

    def _init_local_config(self):
        self.target_dir = self._create_target_directory()
        self.project_config, self.project_config_file_name = self._load_project_config()  # type: ignore
        return

    def _init_dbt_config(self):
        self.dbt_project = self._get_dbt_project_name()
        self.dbt_nodes = self._load_dbt_manifest_nodes()

    @manage_errors
    def apply(
        self,
        apply_namespace: str,
        dry_run: bool = False,
        skip_confirmation: bool = False,
        abort_on_error=True,
        create_non_ingested_tables: bool = False,
    ):
        """
        Compile configuration.
        Submit configuration via API
        """
        _, ns_mc_config_dict, _ = self.compile(apply_namespace, abort_on_error)
        assert self.project_config is not None
        responses = []
        for namespace, mc_config_dict in ns_mc_config_dict.items():
            self._print_func("----------------------------------------------------------------")
            self._print_func(f"Computing changes for namespace: {namespace}")
            self._print_func("----------------------------------------------------------------")
            response = self.client.apply_config_template(
                namespace,
                mc_config_dict,
                resource=self.project_config.default_resource,
                dry_run=True,
                misconfigured_as_warning=True,
                create_non_ingested_tables=create_non_ingested_tables,
            )
            self._check_apply_errors(response, abort_on_error)  # type: ignore

            if response.warnings:  # type: ignore
                click.echo(click.style("WARNINGS encountered, please review:", fg="red", bold=True))
                click.echo()

                # Global warnings should be shown first.
                # Monitor specific warnings should be shown next.
                for name, warnings in response.warnings.items():  # type: ignore
                    if name == "misconfigured_warnings":
                        continue  # legacy warning that doesn't fit the new format
                    for kind, message in warnings.items():
                        click.echo(f" - {name}: {kind}: {message}")

                misconfigured_warnings = response.warnings.get("misconfigured_warnings", [])  # type: ignore
                if misconfigured_warnings:
                    click.echo("misconfigured monitors:")
                    self._echo_misconfigured_warnings(misconfigured_warnings)

            if response.resource_modifications:  # type: ignore
                self._print_func()
                self._print_func("Modifications:")

                for resource_modification in response.resource_modifications:  # type: ignore
                    header = f" - {'[DRY RUN] ' if dry_run else ''}{resource_modification.type}"
                    self._print_func(
                        f"\n{header}"
                        f"\n{len(header) * '-'}"
                        f"\n{resource_modification.description}"
                        f"\nchanges detected: "
                        f"\n{'_' * 100} "
                        f"\n{resource_modification.diff_string if resource_modification.diff_string else ''}"  # noqa
                        f"{self._generate_significance_text(resource_modification)}"
                        f"\n{'_' * 100}"
                        f"\n\n"
                    )
                self._print_func()

                if dry_run:
                    self._print_func("Dry run, none of the changes have been applied.")
                    self._print_func()
                elif skip_confirmation or click.confirm(
                    "Do you want to apply this change?", abort=False
                ):
                    response = self.client.apply_config_template(
                        namespace,
                        mc_config_dict,
                        resource=self.project_config.default_resource,
                        dry_run=False,
                        misconfigured_as_warning=True,
                        create_non_ingested_tables=create_non_ingested_tables,
                    )
                    success = self._check_apply_errors(response, abort_on_error)  # type: ignore

                    if success:
                        self._print_func("Changes successfully applied.")
                    else:
                        self._print_func("Failed to apply changes.")
                    self._print_func()
                else:
                    self._print_func("None of the changes have been applied.")
            else:
                self._print_func()
                self._print_func("No changes to configuration found, doing nothing.")
                self._print_func()
            responses.append(response)
        return responses

    def _check_apply_errors(
        self,
        response: ConfigTemplateUpdateState,
        abort_on_error: bool,
    ) -> bool:
        if response.state == "SKIPPED":
            self._print_func()
            self._print_func("Update skipped in favor of more recent update.")

            if abort_on_error:
                complain_and_abort("Update skipped, exiting.")

            return False

        if not response.errors:
            return True

        self._print_func()
        self._print_func("Errors encountered when attempting to apply configuration:")

        system_error = response.errors.get("system_error", {})
        if system_error:
            self._print_func(f" - {system_error}")
        else:
            validation_errors = response.errors.get("validation_errors", {})
            try:
                error_json = json.dumps(validation_errors, indent=2)
            except Exception as e:
                error_json = (
                    f"Couldnt parse the validation errors: {e}\nraw errors: {validation_errors}"
                )
            self._print_func(f" - Validation error:\n{error_json}")

        self._print_func()

        if abort_on_error:
            complain_and_abort("Errors encountered, exiting.")

        return False

    @staticmethod
    def _generate_significance_text(resource_modification: ResourceModification):
        if resource_modification.is_significant_change:
            return "\nWARNING: This is a significant change which will reset/retrain the monitor."
        return ""

    @staticmethod
    def _echo_misconfigured_warnings(misconfigured_warnings: List[Dict[str, Any]]):
        for group in misconfigured_warnings:
            click.echo(f" - {group['title']}:")
            for item in group["items"]:
                click.echo(f"     - {item}")

    @manage_errors
    def compile(
        self, apply_namespace, abort_on_error=True
    ) -> Tuple[List[str], Dict[str, Box], Dict[str, Any]]:
        """
        Gather monitor configuration YAML files from project.
        This may also include dbt schema files with meta properties
        """
        self._init_local_config()
        self._print_func()
        self._print_func("Gathering monitor configuration files.")
        assert self.project_config is not None
        files = self._gather_yaml_files(self.project_config)

        if self.project_config.namespace:
            self._print_func(
                f"namespace: {self.project_config.namespace} found in "
                f"{self.project_config_file_name} ignoring value "
                f"passed with --namespace={apply_namespace}"
            )
            apply_namespace = self.project_config.namespace

        if self.dbt_manifest_path:
            self._init_dbt_config()

        errors_by_file = {}
        ns_mc_config_list_map: Dict[str, List[Box]] = {}

        found_config = False

        files = sorted(list(files))
        for file in files:
            if os.stat(file).st_size == 0:
                self._print_func(f" - Skipping {file}, empty file.")
                continue
            with open(file, "r") as f:
                try:
                    yaml_as_dict = Box(yaml.load(f, Loader=TemplateLoader))
                except Exception as e:
                    errors_by_file[file] = [f"Failed to parse YAML: {e}"]
                    self._print_func(f" - Skipping {file}, not a valid YAML file.")
                    continue
            namespace = apply_namespace
            if has_namespace_property(yaml_as_dict):
                # override apply_namespace with namespace in the yaml file.
                namespace = yaml_as_dict.namespace

            if not namespace:
                errors_by_file[file] = [
                    f"A default namespace need to be passed through command line --namespace or "
                    f"set default namespace in {self.project_config_file_name} or "
                    f"an override namespace needs to be set in file: {file}"
                ]
                continue
            if namespace not in ns_mc_config_list_map:
                ns_mc_config_list_map[namespace] = []
            mc_config_list = ns_mc_config_list_map[namespace]

            # This file is a dbt schema file.
            # Let's attempt to parse out special embedded MC sections from meta:
            # See https://docs.getdbt.com/reference/resource-properties/meta
            if is_dbt_schema(yaml_as_dict):
                # need to reload again without the yaml_line and yaml_file_name properties
                with open(file, "r") as reload_file:
                    yaml_as_dict = Box(yaml.safe_load(reload_file))

                for model in yaml_as_dict.models:
                    with contextlib.suppress(KeyError):
                        if model.meta:
                            mc_config = model.meta.montecarlo
                            errors = self._validate_mc_config(mc_config)
                            if self.dbt_manifest_path:
                                mc_config = Box(self._resolve_templates(mc_config))
                            mc_config_list.append(mc_config)
                            found_config = True
                        else:
                            errors = ["meta tag should be removed if its empty"]
                        if errors:
                            errors_by_file[file] = errors
                        self._print_func(f" - {file} - Embedded monitor configuration found.")

            # This file has a root montecarlo property
            # This can be in a standalone file, and also a dbt schema file
            if has_montecarlo_property(yaml_as_dict):
                mc_config = yaml_as_dict.montecarlo
                mc_config.pop("yaml_line", None)
                mc_config.pop("yaml_file_name", None)
                errors = self._validate_mc_config(mc_config)
                if self.dbt_manifest_path:
                    mc_config = Box(self._resolve_templates(mc_config))
                if errors:
                    errors_by_file[file] = errors
                mc_config_list.append(mc_config)
                found_config = True
                self._print_func(f" - {file} - Monitor configuration found.")

        # Exit if there are errors
        if errors_by_file:
            self._print_func()
            self._print_func("Configuration validation errors:")
            for file, errors in errors_by_file.items():
                self._print_func(f" - File: {file}")
                for error in errors:
                    self._print_func(f"    - {error}")
            self._print_func()
            if abort_on_error:
                complain_and_abort("Errors encountered, exiting.")

        if (not found_config) and abort_on_error:
            complain_and_abort(
                "Sorry, we didn't find any YAML files containing Monte Carlo configuration."
            )

        ns_compiled_mc_config: Dict[str, Box] = {}
        for namespace, mc_config_list in ns_mc_config_list_map.items():
            # Merge configs into a single config
            compiled_mc_config = self._merge_mc_configs(mc_config_list)

            # Write configs to target directory. Useful for debugging.
            with open(
                os.path.join(self.target_dir, f"{namespace}_montecarlo_configuration.yml"),
                "w",
            ) as f:
                f.write(compiled_mc_config.to_yaml())

            with open(
                os.path.join(self.target_dir, f"{namespace}_montecarlo_configuration.json"),
                "w",
            ) as f:
                json.dump(compiled_mc_config, f, indent=4)

            ns_compiled_mc_config[namespace] = compiled_mc_config

        return files, ns_compiled_mc_config, errors_by_file

    def _create_target_directory(self) -> str:
        target_dir = os.path.join(self.project_dir, TARGET_DIRECTORY_NAME)
        mkdirs(target_dir)
        return target_dir

    def _load_project_config(self) -> Optional[Tuple[Optional[ProjectConfig], str]]:
        contents = None
        project_config_file_name = PROJECT_CONFIG_FILENAME_YML
        try:
            try:
                with open(os.path.join(self.project_dir, PROJECT_CONFIG_FILENAME_YML), "r") as f:
                    contents = f.read()
            except FileNotFoundError:
                with open(os.path.join(self.project_dir, PROJECT_CONFIG_FILENAME_YAML), "r") as f:
                    contents = f.read()
                project_config_file_name = PROJECT_CONFIG_FILENAME_YAML
        except FileNotFoundError:
            complain_and_abort(
                f"Not a Monte Carlo project. Must define either {PROJECT_CONFIG_FILENAME_YML} "
                f"or {PROJECT_CONFIG_FILENAME_YAML} in current working directory."
            )

        assert contents is not None
        project_config_as_dict = yaml.safe_load(contents)
        try:
            s = ProjectConfig.schema()  # type: ignore
            return s.load(project_config_as_dict), project_config_file_name
        except ValidationError as e:
            complain_and_abort(
                f"Encountered a validation problem in {project_config_file_name}: {e.messages}"
            )

    def _merge_mc_configs(self, mc_config_list: List[Box]) -> Box:
        """
        Merge multiple mc configs into a single one.
        Nested dicts will be merged
        Nested lists will be concatenated
        """

        def merge_configs(x: Box, y: Box):
            # box_merge_lists="extend" does not recurse, so we need to process the
            # first config level ourselves to account for groups (like 'notifications')
            for key, value in y.items():
                if key in x:
                    if isinstance(value, dict):
                        x[key].merge_update(y[key], box_merge_lists="extend")
                    elif isinstance(value, list):
                        x[key].extend(y[key])
                else:
                    x[key] = value

            return x

        compiled_mc_config = reduce(merge_configs, mc_config_list)
        compiled_mc_config.pop(
            "box_merge_lists", None
        )  # Strangely enough, Box adds this add'l key. Remove it
        return compiled_mc_config

    def _validate_mc_config(self, mc_config_dict: Box) -> List[str]:
        """
        Perform simple validation.
        More comprehensive validation is performed server-side
        """
        expected_keys = [
            "field_health",
            "dimension_tracking",
            "json_schema",
            "custom_sql",
            "freshness",
            "volume",
        ]

        errors = []
        for key in expected_keys:
            with contextlib.suppress(KeyError):
                if key in mc_config_dict and not isinstance(mc_config_dict[key], list):
                    errors.append(f'"{key}" property should be a list.')
        return errors

    def _gather_yaml_files(self, project_config: ProjectConfig) -> Set[str]:
        files = set()
        for pattern in project_config.include_file_patterns:  # type: ignore
            for fn in glob.glob(os.path.join(self.project_dir, pattern), recursive=True):
                files.add(fn)
        for pattern in project_config.exclude_file_patterns:  # type: ignore
            excluded = set(fnmatch.filter(files, os.path.join(self.project_dir, pattern)))
            files = files - excluded
        return files

    def _get_dbt_project_name(self) -> Optional[str]:
        dbt_project = os.path.join(self.project_dir, DBT_CONFIG_YML)
        try:
            try:
                with open(dbt_project, "r") as f:
                    return yaml.safe_load(f)["name"]
            except FileNotFoundError:
                dbt_project = os.path.join(self.project_dir, DBT_CONFIG_YAML)
                with open(dbt_project, "r") as f:
                    return yaml.safe_load(f)["name"]
        except Exception:
            complain_and_abort(
                f"Failed to read dbt project name from {DBT_CONFIG_YML} / {DBT_CONFIG_YAML}."
            )

    def _load_dbt_manifest_nodes(self) -> Optional[dict]:
        try:
            with open(self.dbt_manifest_path, "r") as f:  # type: ignore
                return json.load(f)["nodes"]
        except Exception:
            complain_and_abort(f"Failed to load dbt manifest from {self.dbt_manifest_path}.")

    def _resolve_dbt_ref(self, match, project_delimiter: str = ":"):
        model = match.group(1)
        node = self.dbt_nodes.get(f"model.{self.dbt_project}.{model}")  # type: ignore
        if not node:
            complain_and_abort(f"Could not resolve dbt ref for {model}.")

        database = node["database"]  # type: ignore
        schema = node["schema"]  # type: ignore
        name = node.get("alias") or node["name"]  # type: ignore

        return f"{database}{project_delimiter}{schema}.{name}"

    def _resolve_sql_dbt_ref(self, match):
        # We use BQ `:` internally, but other warehouses use `.` so for SQL we need to
        # avoid using `:`. BQ supports `.` as well as `:` so we can just always use `.`.
        return self._resolve_dbt_ref(match, project_delimiter=".")

    def _resolve_templates(self, yaml_object, ref_resolver: Optional[Callable] = None):
        ref_resolver = ref_resolver or self._resolve_dbt_ref

        if isinstance(yaml_object, str):
            return self.DBT_REF_PATTERN.sub(ref_resolver, yaml_object)
        elif isinstance(yaml_object, list):
            return [self._resolve_templates(i, ref_resolver) for i in yaml_object]
        elif isinstance(yaml_object, dict):
            return {
                k: self._resolve_templates(
                    v,
                    self._resolve_sql_dbt_ref if k in self.SQL_KEYS else ref_resolver,
                )
                for k, v in yaml_object.items()
            }

        return yaml_object

    @manage_errors
    def delete(self, namespace: str, dry_run: bool = False, abort_on_error=True):
        """
        Delete all monitors in namespace
        """
        response = self.client.delete_config_template(namespace, dry_run=dry_run)

        dry_run_s = "[DRY RUN] " if dry_run else ""
        self._print_func(
            f"\n{dry_run_s}Deleted {response.num_deleted} resources in namespace={namespace}\n"
        )

        return response

    @manage_errors
    def list_namespaces(self, limit=100):
        """
        Get all namespaces
        """
        query = Query()
        templates = query.get_monte_carlo_config_templates(first=limit + 1)
        assert templates.edges is not None
        templates.edges.node.__fields__("namespace", "last_update_time")
        response = self._pycarlo_client(
            query,
            additional_headers={
                "x-mcd-telemetry-reason": "cli",
                "x-mcd-telemetry-service": "iac_service",
                "x-mcd-telemetry-command": self._command_name,
            },
        )
        namespaces = [
            (edge.node.namespace, edge.node.last_update_time)
            for edge in response.get_monte_carlo_config_templates.edges
            if edge.node is not None
        ]

        more_ns_available = False
        if len(namespaces) > limit:
            namespaces = namespaces[:-1]
            more_ns_available = True
        self._print_func(tabulate(namespaces, headers=self.NS_HEADERS, tablefmt="fancy_grid"))

        if more_ns_available:
            self._print_func(self.MORE_NS_MESSAGE)

    @manage_errors
    def generate_from_dbt_tests(
        self,
        output_path: str,
        test_types: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
    ):
        """
        Create Monte Carlo monitor config YAML from dbt tests in dbt manifest.
        May specify a subset of dbt test types to import.
        """
        tests = (
            n
            for n in self._load_dbt_manifest_nodes().values()  # type: ignore
            if n.get("resource_type") == "test"
        )

        if test_types:
            test_type_set = set(test_types)
            tests = (
                t
                for t in tests
                if t.get("test_metadata", {}).get("name", "SINGULAR") in test_type_set
                # Singular tests do not have metadata
            )

        common_args = {
            "schedule": {
                "type": "fixed",
                "start_time": datetime.now(timezone.utc).isoformat(),
                "interval_minutes": 720,
            },
            "comparisons": [{"type": "threshold", "operator": "GT", "threshold_value": 0}],
        }

        if labels:
            common_args["labels"] = sorted(list(set(labels)))

        custom_sql_rules = [
            {
                "name": t["unique_id"],
                "description": t["unique_id"],
                "sql": t["compiled_code"].strip(),
                **common_args,
            }
            for t in tests
        ]

        self._print_func(f"Found {len(custom_sql_rules)} dbt tests.")

        config = {"montecarlo": {"custom_sql": custom_sql_rules}}
        with open(output_path, "w", encoding="utf-8", errors="strict") as f:
            yaml.dump(
                Box(config).to_dict(),
                stream=f,
                default_flow_style=False,
                sort_keys=False,
                width=float("inf"),
            )

        self._print_func(f"Wrote monitor config to {output_path}.")

    @manage_errors
    def convert_to_ui(self, namespace: str, dry_run: bool = False):
        if not dry_run:
            if not click.confirm(
                "Do you want to convert your monitor as code to UI monitors? "
                "If you proceed, your monitors will become editable through the UI and "
                "applying the monitor configuration "
                "will create new monitors.",
                abort=False,
            ):
                return
        convert_to_ui = """
                        mutation convertConfigTemplateToUiMonitors(
                            $namespace: String!
                            $dryRun: Boolean
                        ) {
                          convertConfigTemplateToUiMonitors(
                            namespace: $namespace
                            dryRun: $dryRun
                          ) {
                            response {
                              monitors {
                                uuid
                                name
                              }
                            }
                          }
                        }
                    """
        response = self._pycarlo_client(
            convert_to_ui,
            {"namespace": namespace, "dryRun": dry_run},
            additional_headers={
                "x-mcd-telemetry-reason": "cli",
                "x-mcd-telemetry-service": "iac_service",
                "x-mcd-telemetry-command": self._command_name,
            },
        )

        monitors = response.convert_config_template_to_ui_monitors.response.monitors  # type: ignore
        if not monitors:
            complain_and_abort(
                "No monitors converted to UI. Check that the namespace exists and is not empty."
            )
            return

        self._print_func("Exported the following monitors to UI:")
        for monitor in response.convert_config_template_to_ui_monitors.response.monitors:  # type: ignore
            self._print_func(f"* {monitor.uuid}: {monitor.name}")

    @manage_errors
    def convert_to_mac(
        self,
        namespace: str,
        project_dir: str,
        monitors_file: Optional[str] = None,
        all_monitors: bool = False,
        dry_run: bool = False,
    ):
        if all_monitors and monitors_file:
            complain_and_abort(
                "You cannot export all monitors and provide a monitors file at the same time."
            )

        if monitors_file:
            monitors_uuids = self._read_monitors_uuids_from_file(monitors_file)
        else:
            monitors_uuids = None

        if not dry_run:
            if not click.confirm(
                "Do you want to convert your monitors to monitors as code? "
                "If you proceed, your monitors will no longer be editable through the UI and "
                "you'll need to edit and apply the monitor configuration to perform changes.",
                abort=False,
            ):
                return
        convert_to_mac = """
                mutation convertUiMonitorsToConfigTemplate(
                    $namespace: String!
                    $monitorUuids: [UUID]
                    $allMonitors: Boolean
                    $dryRun: Boolean
                ) {
                  convertUiMonitorsToConfigTemplate(
                    namespace: $namespace
                    monitorUuids: $monitorUuids
                    allMonitors: $allMonitors
                    dryRun: $dryRun
                  ) {
                    response {
                      configTemplateAsDict
                      errors
                    }
                  }
                }
            """

        response = self._pycarlo_client(
            convert_to_mac,
            {
                "namespace": namespace,
                "monitorUuids": ([str(m) for m in monitors_uuids] if monitors_uuids else None),
                "allMonitors": all_monitors,
                "dryRun": dry_run,
            },
            additional_headers={
                "x-mcd-telemetry-reason": "cli",
                "x-mcd-telemetry-service": "iac_service",
                "x-mcd-telemetry-command": self._command_name,
            },
        )

        graphql_errors = getattr(response, "errors", None)
        if graphql_errors:
            graphql_errors_str = "\n".join(graphql_errors)
            complain_and_abort(f"Errors encountered during export:\n{graphql_errors_str}")
            return

        errors = response.convert_ui_monitors_to_config_template.response.errors  # type: ignore
        if errors:
            errors_str = "\n".join(errors)  # type: ignore
            complain_and_abort(f"Errors encountered during export:\n{errors_str}")
            return

        self._export_project(
            project_dir,
            namespace,
            response.convert_ui_monitors_to_config_template.response.config_template_as_dict,  # type: ignore
        )

    @manage_errors
    def export_as_latest(self, namespace: str, monitors_file: Optional[str] = None):
        if monitors_file:
            monitor_uuids = self._read_monitors_uuids_from_file(monitors_file)
        else:
            monitor_uuids = []

        export_as_latest = """
            query exportLatestVersionMonitorReplacementTemplates(
                $namespace: String!
                $monitorUuids: [UUID!]!
            ) {
              exportLatestVersionMonitorReplacementTemplates(
                namespace: $namespace
                monitorUuids: $monitorUuids
                validate: false
              ) {
                configTemplateAsYaml
                errors
              }
            }
        """
        response = self._pycarlo_client(
            export_as_latest,
            {"namespace": namespace, "monitorUuids": [str(m) for m in monitor_uuids]},
            additional_headers={
                "x-mcd-telemetry-reason": "cli",
                "x-mcd-telemetry-service": "iac_service",
                "x-mcd-telemetry-command": self._command_name,
            },
        )
        export_result = response.export_latest_version_monitor_replacement_templates  # type: ignore
        errors = export_result.errors
        if errors:
            complain_and_abort(f"There were errors exporting the monitors: {errors}")
            return

        self._print_func(export_result.config_template_as_yaml)

    @manage_errors
    def get_template(self, namespace: str, project_dir: str):
        get_template = """
                    query getMonteCarloConfigTemplates(
                        $namespace: String!
                    ) {
                      getMonteCarloConfigTemplates(
                        namespace: $namespace
                        first: 1
                      ) {
                        edges {
                          node {
                            template
                          }
                        }
                      }
                    }
                """
        response = self._pycarlo_client(
            get_template,
            {
                "namespace": namespace,
            },
            additional_headers={
                "x-mcd-telemetry-reason": "cli",
                "x-mcd-telemetry-service": "iac_service",
                "x-mcd-telemetry-command": self._command_name,
            },
        )

        graphql_errors = getattr(response, "errors", None)
        if graphql_errors:
            graphql_errors_str = "\n".join(graphql_errors)
            complain_and_abort(f"Errors encountered during export:\n{graphql_errors_str}")
            return

        if not response.get_monte_carlo_config_templates.edges:  # type: ignore
            complain_and_abort(f"Did not found a template for namespace '{namespace}'.")
            return

        self._export_project(
            project_dir,
            namespace,
            response.get_monte_carlo_config_templates.edges[0].node.template,  # type: ignore
        )

    def _export_project(self, project_dir: str, namespace: str, template_json: str):
        template = json.loads(template_json)
        config = {"montecarlo": template}
        if project_dir == "-":
            yaml.dump(
                config,
                stream=sys.stdout,
                default_flow_style=False,
                sort_keys=False,
                width=float("inf"),
            )
        else:
            if os.path.exists(project_dir):
                complain_and_abort(f"{project_dir} already exists.")
                return

            os.makedirs(os.path.join(project_dir))
            with open(
                os.path.join(project_dir, "montecarlo.yml"),
                "w",
                encoding="utf-8",
                errors="strict",
            ) as f:
                f.writelines(
                    f"""version: 1
namespace: {namespace}
"""
                )
            os.mkdir(os.path.join(project_dir, "montecarlo"))

            with open(
                os.path.join(project_dir, "montecarlo/monitors.yml"),
                "w",
                encoding="utf-8",
                errors="strict",
            ) as f:
                yaml.dump(
                    config,
                    stream=f,
                    default_flow_style=False,
                    sort_keys=False,
                    width=float("inf"),
                )

            self._print_func(f"Wrote monitor config to {project_dir}.")

    @staticmethod
    def _str_presenter(dumper, data):
        # Configures yaml for dumping multiline strings
        if data.count("\n") > 0:
            # It is a multiline string. Use a | style
            return dumper.represent_scalar(
                tag="tag:yaml.org,2002:str",
                value=MonteCarloConfigService._prepare_multiline_str(data),
                style="|",
            )

        return dumper.represent_scalar("tag:yaml.org,2002:str", data)

    @staticmethod
    def _prepare_multiline_str(data: str) -> str:
        """
        Prepare the data to format as multiline string in the YAML. It must meet
        some conditions to represent it as a block by yaml.dump,
        """

        # Trailing spaces are not permitted (check analyze_scalar in emitter.py,
        # and how allow_block is defined)
        result = "\n".join([line.rstrip() for line in data.splitlines()])
        if not result.endswith("\n"):
            # Add a line break at the end (to avoid using |- and use | in the yaml)
            result += "\n"

        return result

    @staticmethod
    def _read_monitors_uuids_from_file(monitors_file: str) -> List[UUID]:
        if not os.path.exists(monitors_file):
            complain_and_abort(f"File {monitors_file} does not exist.")

        monitor_uuids = []
        with open(monitors_file, "r") as f:
            line_number = 0
            for line in f.readlines():
                line_number += 1
                line = line.strip()
                if line:
                    try:
                        monitor_uuids.append(UUID(line))
                    except Exception:
                        complain_and_abort(f"Invalid UUID in monitors file on line {line_number}")

        if not monitor_uuids:
            complain_and_abort(f"File {monitors_file} does not contain any valid monitors.")

        return monitor_uuids



from datetime import date

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Monte Carlo's CLI"
copyright = f"{date.today().year}, Monte Carlo Data, Inc"
author = "Monte Carlo Data, Inc"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx_click"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]

html_theme_options = {
    "external_links": [
        {"name": "Dashboard", "url": "https://getmontecarlo.com"},
        {"name": "API reference", "url": "https://apidocs.getmontecarlo.com/"},
        {"name": "Product docs", "url": "https://docs.getmontecarlo.com/"},
    ],
    "favicons": [
        {
            "rel": "icon",
            "sizes": "16x16",
            "href": "logo.png",
        }
    ],
    "logo": {
        "text": "Monte Carlo's CLI Reference",
    },
    "search_bar_text": "Search",
    "show_toc_level": 2,
}

html_context = {"default_mode": "dark"}



import click

from montecarlodata.insights.data_insights import InsightsService
from montecarlodata.insights.fields import (
    CLEANUP_SUGGESTIONS_INSIGHT_NAME,
    COVERAGE_OVERVIEW_INSIGHT_NAME,
    DETERIORATING_QUERY_INSIGHT_NAME,
    EVENTS_INSIGHT_NAME,
    FILE_SCHEME,
    INCIDENT_QUERY_INSIGHT_NAME,
    KEY_ASSETS_INSIGHT_NAME,
    READ_WRITE_INSIGHT_NAME,
    RULES_AND_SLIS_INSIGHT_NAME,
    S3_SCHEME,
    SCHEME_DELIM,
)
from montecarlodata.tools import add_common_options

# Shared command verbiage
GET_VERBIAGE = f"""
\b\n
DESTINATION is the path where the insight will be written to.

Supported schemes:\n
    '{FILE_SCHEME}{SCHEME_DELIM}' - save insight locally.\n
    '{S3_SCHEME}{SCHEME_DELIM}' - save insight to S3.

Notice - Will overwrite a file if it exists in the path and create any missing directories or prefixes.
"""  # noqa: E501

# Options shared across commands
MINIMAL_GET_OPTIONS = [
    click.option(
        "--aws-profile",
        required=False,
        help="AWS profile to be used when uploading to S3.",
    ),
    click.option(
        "--dry",
        required=False,
        is_flag=True,
        help="Echo temporary presigned URL for the insight and quit.",
    ),
]

GET_OPTIONS = [click.argument("destination", required=True), *MINIMAL_GET_OPTIONS]


@click.group(help="Aggregated insights on your tables.")
def insights():
    """
    Group for any insight related subcommands
    """
    pass


@insights.command(help="List insights details and availability.", name="list")
@click.pass_obj
def list_insights(ctx):
    InsightsService(
        config=ctx["config"],
        command_name="insights list_insights",
    ).echo_insights()


@insights.command(help=f"Get key assets insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_key_assets(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_key_assets",
    ).get_insight(insight=KEY_ASSETS_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get coverage overview (monitors) insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_coverage_overview(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_coverage_overview",
    ).get_insight(insight=COVERAGE_OVERVIEW_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get cleanup suggestions insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_cleanup_suggestions(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_cleanup_suggestions",
    ).get_insight(insight=CLEANUP_SUGGESTIONS_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get events insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_events(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_events",
    ).get_insight(insight=EVENTS_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get table read/write activity insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_table_activity(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_table_activity",
    ).get_insight(insight=READ_WRITE_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get incident query changes insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_incident_queries(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_incident_queries",
    ).get_insight(insight=INCIDENT_QUERY_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get deteriorating queries insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_deteriorating_queries(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_deteriorating_queries",
    ).get_insight(insight=DETERIORATING_QUERY_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get rule and SLI results insight. {GET_VERBIAGE}")
@add_common_options(GET_OPTIONS)
@click.pass_obj
def get_rule_results(ctx, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get_rule_results",
    ).get_insight(insight=RULES_AND_SLIS_INSIGHT_NAME, **kwargs)


@insights.command(help=f"Get an arbitrary insight. {GET_VERBIAGE}")
@add_common_options(MINIMAL_GET_OPTIONS)
@click.option(
    "--name",
    help="Name of the insight to retrieve. Insight names can be found "
    'via the list command (e.g. `key_assets` from "Key Assets (key_assets)")',
    required=True,
)
@click.option(
    "--destination",
    help="Destination location to save insight.",
    required=True,
)
@click.pass_obj
def get(ctx, name, **kwargs):
    InsightsService(
        config=ctx["config"],
        command_name="insights get",
    ).get_insight(insight=name, **kwargs)



from typing import Optional

import click

from montecarlodata.collector.fields import (
    EXPECTED_TTC_RESPONSE_FIELD,
    EXPECTED_TTOC_RESPONSE_FIELD,
)
from montecarlodata.config import Config
from montecarlodata.errors import manage_errors
from montecarlodata.queries.collector import (
    TEST_TCP_OPEN_CONNECTION,
    TEST_TELNET_CONNECTION,
)
from montecarlodata.utils import GqlWrapper


class CollectorNetworkTestService:
    def __init__(
        self,
        config: Config,
        command_name: str,
        request_wrapper: Optional[GqlWrapper] = None,
    ):
        self._abort_on_error = True
        self._command_name = command_name

        self._request_wrapper = request_wrapper or GqlWrapper(
            config,
            command_name=self._command_name,
        )

    @manage_errors
    def echo_telnet_test(self, **kwargs):
        """
        Checks if telnet connection is usable and echos results
        """
        self._echo_network_test_validations(
            query=TEST_TELNET_CONNECTION, operation=EXPECTED_TTC_RESPONSE_FIELD, **kwargs
        )

    @manage_errors
    def echo_tcp_open_test(self, **kwargs):
        """
        Checks if a destination exists and accepts requests and echos results
        """
        self._echo_network_test_validations(
            query=TEST_TCP_OPEN_CONNECTION, operation=EXPECTED_TTOC_RESPONSE_FIELD, **kwargs
        )

    def _echo_network_test_validations(self, query: str, operation: str, **kwargs):
        """
        Do request and echos a validation response
        """
        response = self._request_wrapper.make_request_v2(
            query=query,
            operation=operation,
            service="collector_network_test",
            variables=kwargs,
        )
        click.echo(response.data.validations[0].message)  # type: ignore



import copy
from unittest import TestCase
from unittest.mock import Mock, call, patch

import click
from box import Box

from montecarlodata.common.data import AwsGlueAthenaResourceProperties
from montecarlodata.common.resources import CloudResourceService
from montecarlodata.common.user import UserService
from montecarlodata.discovery.policy_gen import PolicyDiscoveryService
from montecarlodata.utils import GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG
from tests.test_network_discovery import SAMPLE_DC_PROPS

SAMPLE_BUCKET = "foo"
SAMPLE_PATH = "/bar/qux"
SAMPLE_GET_TABLES = {
    "TableList": [{"StorageDescriptor": {"Location": f"s3://{SAMPLE_BUCKET}{SAMPLE_PATH}"}}],
    "NextToken": None,
}
SAMPLE_GET_WORKGROUP = {
    "Configuration": {
        "ResultConfiguration": {"OutputLocation": f"s3://{SAMPLE_BUCKET}{SAMPLE_PATH}"}
    }
}


class PolicyDiscoveryTest(TestCase):
    def setUp(self) -> None:
        self.maxDiff = None
        self._cloud_resource_service_mock = Mock(autospec=CloudResourceService)
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)

        self._service = PolicyDiscoveryService(
            config=_SAMPLE_CONFIG,
            command_name="test",
            cloud_resource_service=self._cloud_resource_service_mock,
            user_service=self._user_service_mock,
            request_wrapper=self._request_wrapper_mock,
        )

    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.generate_token")
    @patch("montecarlodata.discovery.policy_gen.read_files")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_cf_role(self, render_mock, read_mock, generate_mock, click_mock):
        account, external_id, file_vals, template_vals = (
            "foo",
            "1234",
            {"foo": "bar"},
            {"template": {"foo": "bar"}},
        )
        files, collector_props = (
            ["a", "b", "c"],
            Box({"stackArn": f"arn:partition:service:region:{account}:resource"}),
        )

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        generate_mock.return_value = external_id
        read_mock.return_value = file_vals
        render_mock.return_value = template_vals

        self._service.generate_cf_role(policy_files=files)

        read_mock.assert_called_once_with(files)
        generate_mock.assert_called_once_with()
        render_mock.assert_called_once_with(
            path="../templates/aws_role.json",
            collector_account_id=account,
            collector_external_id=external_id,
            resource_policies=file_vals,
        )
        click_mock.assert_has_calls(
            [
                call.echo(
                    "Generating CloudFormation template for creating a role resource.",
                    err=True,
                ),
                call.echo(template_vals),
            ]
        )

    @patch("montecarlodata.discovery.policy_gen.complain_and_abort")
    def test_generate_cf_role_without_a_policy(self, complain_mock):
        complain_mock.side_effect = click.exceptions.Abort

        with self.assertRaises(click.exceptions.Abort):
            self._service.generate_cf_role(policy_files=[])
        complain_mock.assert_called_once_with("At least one policy is required.")

    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_msk_policy(self, render_mock, click_mock):
        # given
        options, _, _, account_id = (
            {"foo": "bar"},
            {"a", "b"},
            {"c", "d"},
            "1234",
        )
        rendering = {"hello": "world"}
        dc_props = copy.deepcopy(SAMPLE_DC_PROPS)
        dc_props.resources_client.get_caller_identity.return_value = account_id

        self._cloud_resource_service_mock.get_dc_resource_props.return_value = dc_props

        render_mock.return_value = rendering

        # when
        self._service.generate_msk_policy(**options)

        # then
        render_mock.assert_called_once_with(
            path="../templates/msk_policy.json",
            region="us-east-1",
            account_id=account_id,
        )
        click_mock.assert_has_calls(
            [call.echo("Generating policy for MSK.", err=True), call.echo(rendering)]
        )

    @patch.object(PolicyDiscoveryService, "_get_common_glue_athena_props")
    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_glue_policy(self, render_mock, click_mock, common_mock):
        options, database_name, bucket_names, account_id = (
            {"foo": "bar"},
            {"a", "b"},
            {"c", "d"},
            "1234",
        )
        rendering = {"hello": "world"}
        props = copy.deepcopy(SAMPLE_DC_PROPS)

        common_mock.return_value = AwsGlueAthenaResourceProperties(
            database_names=database_name,
            bucket_names=bucket_names,
            dc_resource_props=props,
        )
        props.resources_client.get_caller_identity.return_value = account_id
        render_mock.return_value = rendering

        self._service.generate_glue_policy(**options)

        common_mock.assert_called_once_with(**options)
        render_mock.assert_called_once_with(
            path="../templates/glue_policy.json",
            region="us-east-1",
            account_id=account_id,
            database_names=database_name,
            data_buckets=bucket_names,
        )
        click_mock.assert_has_calls(
            [call.echo("Generating policy for Glue.", err=True), call.echo(rendering)]
        )

    @patch.object(PolicyDiscoveryService, "_get_workgroup_results")
    @patch.object(PolicyDiscoveryService, "_get_common_glue_athena_props")
    @patch("montecarlodata.discovery.policy_gen.click")
    @patch("montecarlodata.discovery.policy_gen.render_dumped_json")
    def test_generate_athena_policy(self, render_mock, click_mock, common_mock, workgroup_mock):
        workgroup_name = "workgroup"
        options, bucket_name, path = (
            {"foo": "bar", "workgroup_name": workgroup_name},
            "bucket",
            "prefix/key",
        )
        database_name, bucket_names, account_id, rendering = (
            {"a", "b"},
            {"c", "d"},
            "1234",
            {"hello": "world"},
        )
        props = copy.deepcopy(SAMPLE_DC_PROPS)

        common_mock.return_value = AwsGlueAthenaResourceProperties(
            database_name, bucket_names, props
        )
        workgroup_mock.return_value = (bucket_name, path)
        props.resources_client.get_caller_identity.return_value = account_id
        render_mock.return_value = rendering

        self._service.generate_athena_policy(**options)

        common_mock.assert_called_once_with(foo="bar")
        workgroup_mock.assert_called_once_with(
            workgroup_name=workgroup_name, dc_resource_props=props
        )
        render_mock.assert_called_once_with(
            path="../templates/athena_policy.json",
            region="us-east-1",
            account_id=account_id,
            database_names=database_name,
            data_buckets=bucket_names,
            result_bucket=bucket_name,
            result_path=path,
            workgroup_name=workgroup_name,
        )
        click_mock.assert_has_calls(
            [call.echo("Generating policy for Athena.", err=True), call.echo(rendering)]
        )

    @patch.object(PolicyDiscoveryService, "_get_buckets")
    def test_get_common_glue_athena_props(self, bucket_mock):
        database_names, bucket_names = ["a", "b", "*"], ["c", "d", "d"]
        collector_props, dc_resource_props = Box({"foo": "bar"}), copy.deepcopy(SAMPLE_DC_PROPS)

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = dc_resource_props

        self.assertEqual(
            self._service._get_common_glue_athena_props(
                database_names=database_names, bucket_names=bucket_names
            ),
            AwsGlueAthenaResourceProperties(
                database_names={"*"},
                bucket_names={"c", "d"},
                dc_resource_props=dc_resource_props,
            ),
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(
            dc_id=None
        )
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False,
        )
        bucket_mock.assert_not_called()

    @patch.object(PolicyDiscoveryService, "_get_buckets")
    def test_get_common_glue_athena_props_agent_id(self, bucket_mock):
        database_names, bucket_names = ["a", "b", "*"], ["c", "d", "d"]
        collector_props, dc_resource_props = Box({"foo": "bar"}), copy.deepcopy(SAMPLE_DC_PROPS)
        agent_id = "1234"

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = dc_resource_props

        self._user_service_mock.get_agent.return_value = {"dc_id": "5678"}

        self.assertEqual(
            self._service._get_common_glue_athena_props(
                database_names=database_names,
                bucket_names=bucket_names,
                agent_id=agent_id,
            ),
            AwsGlueAthenaResourceProperties(
                database_names={"*"},
                bucket_names={"c", "d"},
                dc_resource_props=dc_resource_props,
            ),
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(
            dc_id="5678"
        )
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False,
        )
        bucket_mock.assert_not_called()

    @patch.object(PolicyDiscoveryService, "_get_buckets")
    def test_get_common_glue_athena_props_with_no_buckets(self, bucket_mock):
        database_names, bucket_names, dc_id = ["a", "b", "*"], {"c", "d"}, "1234"
        collector_props, dc_resource_props = Box({"foo": "bar"}), copy.deepcopy(SAMPLE_DC_PROPS)

        bucket_mock.return_value = bucket_names

        self._cloud_resource_service_mock.get_and_validate_active_collector.return_value = (
            collector_props
        )
        self._cloud_resource_service_mock.get_dc_resource_props.return_value = dc_resource_props

        self.assertEqual(
            self._service._get_common_glue_athena_props(database_names=database_names, dc_id=dc_id),
            AwsGlueAthenaResourceProperties(
                database_names={"*"},
                bucket_names=bucket_names,
                dc_resource_props=dc_resource_props,
            ),
        )
        self._cloud_resource_service_mock.get_and_validate_active_collector.assert_called_once_with(
            dc_id=dc_id
        )
        self._cloud_resource_service_mock.get_dc_resource_props.assert_called_once_with(
            collector_props=collector_props,
            get_stack_outputs=False,
            get_stack_params=False,
        )
        bucket_mock.assert_called_once_with(
            database_names={"*"}, dc_resource_props=dc_resource_props
        )

    @patch("montecarlodata.discovery.policy_gen.click.progressbar")
    @patch.object(PolicyDiscoveryService, "_get_table_locations")
    def test_get_buckets(self, table_location_mock, click_mock):
        class MockContext:
            def __enter__(self):
                return self

            def __iter__(self):
                for x in database_names:
                    yield x

            def __exit__(self, *args):
                pass

        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        database_names, locations = {"foo", "bar"}, {"bucket1", "bucket2"}
        dc_resource_props = copy.deepcopy(SAMPLE_DC_PROPS)

        table_location_mock.return_value = locations
        click_mock.return_value = MockContext()

        dc_resource_props.resources_client.get_glue_tables.return_value = tables

        self.assertEqual(
            self._service._get_buckets(
                database_names=database_names, dc_resource_props=dc_resource_props
            ),
            locations,
        )
        table_location_mock.assert_has_calls(
            [call(tables=tables.table_list), call(tables=tables.table_list)]
        )
        dc_resource_props.resources_client.get_glue_tables.assert_has_calls(
            [
                call(database_name="bar", page_token=None),
                call(database_name="foo", page_token=None),
            ],
            any_order=True,
        )

    def test_handle_wildcards(self):
        self.assertEqual(self._service._handle_wildcards(None), set())
        self.assertEqual(self._service._handle_wildcards([]), set())
        self.assertEqual(self._service._handle_wildcards(["foo", "foo", "foo"]), {"foo"})
        self.assertEqual(self._service._handle_wildcards(["foo", "foo", "*"]), {"*"})

    @patch("montecarlodata.discovery.policy_gen.click.progressbar")
    def test_get_workgroup_results(self, click_mock):
        workgroup_name, dc_resource_props = "foo", copy.deepcopy(SAMPLE_DC_PROPS)
        workgroup_details = Box(SAMPLE_GET_WORKGROUP, camel_killer_box=True)

        dc_resource_props.resources_client.get_athena_workgroup.return_value = workgroup_details
        self.assertEqual(
            self._service._get_workgroup_results(
                workgroup_name, dc_resource_props=dc_resource_props
            ),
            (SAMPLE_BUCKET, SAMPLE_PATH),
        )
        dc_resource_props.resources_client.get_athena_workgroup.assert_called_once_with(
            workgroup=workgroup_name
        )

    def test_get_table_locations(self):
        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        self.assertEqual(
            list(self._service._get_table_locations(tables=tables.table_list)),
            [SAMPLE_BUCKET],
        )

    def test_get_table_locations_with_invalid_schemes(self):
        tables = Box(SAMPLE_GET_TABLES, camel_killer_box=True)
        tables.table_list[0].storage_descriptor.location = "invalid"
        self.assertEqual(list(self._service._get_table_locations(tables=tables.table_list)), [])



"""
Common commands
"""

import click

from montecarlodata.tools import (
    AdvancedOptions,
    convert_uuid_callback,
)

DC_ID_OPTION = [
    click.option(
        "--collector-id",
        "dc_id",
        required=False,
        type=click.UUID,
        callback=convert_uuid_callback,
        help="ID for the data collector. To disambiguate accounts with multiple collectors.",
        cls=AdvancedOptions,
    ),
]

DISAMBIGUATE_DC_OPTIONS = [
    click.option(
        "--agent-id",
        "agent_id",
        required=False,
        type=click.UUID,
        callback=convert_uuid_callback,
        help="ID for the agent. To disambiguate accounts with multiple agents.",
        cls=AdvancedOptions,
        mutually_exclusive_options=["dc_id"],
    ),
    click.option(
        "--collector-id",
        "dc_id",
        required=False,
        type=click.UUID,
        callback=convert_uuid_callback,
        help="ID for the data collector. To disambiguate accounts with multiple collectors.",
        cls=AdvancedOptions,
        mutually_exclusive_options=["agent_id"],
    ),
]

DC_RESOURCE_OPTIONS = [
    click.option(
        "--resource-aws-region",
        required=False,
        help="Override the AWS region where the resource is located. "
        "Defaults to the region where the collector is hosted.",
    ),
    click.option(
        "--resource-aws-profile",
        required=True,
        help="AWS profile for the resource.",
    ),
]



# Query responses
EXPECTED_GET_DATA_EXPORT_URL = "getDataExportUrl"

# Data Export Names
MONITORS_DATA_EXPORT_NAME = "MONITORS"
ASSETS_DATA_EXPORT_NAME = "ASSETS"
ALERTS_DATA_EXPORT_NAME = "ALERTS"
EVENTS_DATA_EXPORT_NAME = "EVENTS"

# Supported Schemes
FILE_SCHEME = "file"
S3_SCHEME = "s3"
SCHEME_DELIM = "://"

# Data Export headers
LIST_DATA_EXPORTS_HEADERS = ["Data Export (Name)", "Description"]



import uuid
from typing import Optional

import click

from montecarlodata.common import create_mc_client
from montecarlodata.platform.service import PlatformService


@click.group(help="Manage platform settings.")
def platform():
    """
    Group for any platform related subcommands
    """
    pass


@platform.command(help="Tests if all connections can be migrated to the Monte Carlo Platform.")
@click.option(
    "--service-id",
    help="ID for the service as listed by 'platform list' command. "
    "To disambiguate accounts with multiple services.",
    required=False,
)
@click.pass_obj
def test_migration(ctx, service_id: Optional[str] = None):
    PlatformService(
        mc_client=create_mc_client(ctx),
        command_name="platform test_migration",
    ).caas_migration_test(dc_id=uuid.UUID(service_id) if service_id else None)


@platform.command(name="list", help="Lists all services in the account.")
@click.pass_obj
def list_services(ctx):
    PlatformService(
        mc_client=create_mc_client(ctx),
        command_name="platform list_services",
    ).list_services()



from typing import Optional

import click

from montecarlodata.agents.agent import AgentService
from montecarlodata.agents.fields import (
    AWS,
    AWS_ASSUMABLE_ROLE,
    AWS_GENERIC,
    AZURE,
    AZURE_BLOB,
    AZURE_FUNCTION_APP_KEY,
    AZURE_STORAGE_ACCOUNT_KEYS,
    DATA_STORE_AGENT,
    GCP,
    GCP_JSON_SERVICE_ACCOUNT_KEY,
    GCS,
    OAUTH2_CLIENT_CREDENTIALS,
    OAUTH2_PASSWORD,
    REMOTE_AGENT,
    S3,
)
from montecarlodata.collector.commands import NETWORK_TEST_OPTIONS
from montecarlodata.collector.network_tests import CollectorNetworkTestService
from montecarlodata.common import create_mc_client
from montecarlodata.common.commands import DC_ID_OPTION
from montecarlodata.integrations.commands import PASSWORD_VERBIAGE
from montecarlodata.tools import (
    AdvancedOptions,
    add_common_options,
    validate_json_callback,
)

DRY_RUN_OPTIONS = [
    click.option(
        "--dry-run",
        required=False,
        default=False,
        show_default=True,
        is_flag=True,
        help="Dry run (validates credentials but doesn't create agent).",
    ),
]

MUTUAL_TLS_OPTIONS = [
    click.option(
        "--client-certificate",
        help="Path to the file that contains a client certificate for agent connectivity.",
        required=False,
        type=click.Path(dir_okay=False, exists=True),
        cls=AdvancedOptions,
    ),
    click.option(
        "--private-key",
        help="Path to the file that contains a private key for the "
        "client certificate for agent connectivity.",
        required=False,
        type=click.Path(dir_okay=False, exists=True),
        cls=AdvancedOptions,
        required_with_options=["client_certificate"],
    ),
]

UPDATE_AGENT_OPTIONS = [
    click.option(
        "--agent-id",
        help="UUID of existing Agent to update.",
        required=False,
    )
]


@click.group(help="Manage a Monte Carlo Agent.")
def agents():
    """
    Group for any Agent related subcommands
    """
    pass


@agents.command(help="Register a Data Store Agent with remote Azure Blob storage container.")
@click.pass_obj
@click.option(
    "--connection-string",
    help=f"A connection string to an Azure Storage account. {PASSWORD_VERBIAGE}",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option(
    "--container-name",
    help="Name of Azure Storage container for data store.",
    required=True,
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_azure_blob_store(ctx, container_name, **kwargs):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_azure_blob_store",
    ).create_agent(
        agent_type=DATA_STORE_AGENT,
        platform=AZURE,
        storage=AZURE_BLOB,
        auth_type=AZURE_STORAGE_ACCOUNT_KEYS,
        endpoint=container_name,
        **kwargs,
    )


@agents.command(help="Register a Data Store Agent with remote S3 bucket.")
@click.pass_obj
@click.option(
    "--assumable-role",
    help="ARN of AWS assumable role.",
    required=True,
)
@click.option(
    "--bucket-name",
    help="Name of S3 bucket for data store.",
    required=True,
)
@click.option(
    "--external-id",
    help="AWS External ID.",
    required=False,
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_s3_store(ctx, bucket_name, **kwargs):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_s3_store",
    ).create_agent(
        agent_type=DATA_STORE_AGENT,
        platform=AWS,
        storage=S3,
        auth_type=AWS_ASSUMABLE_ROLE,
        endpoint=bucket_name,
        **kwargs,
    )


@agents.command(help="Register a Remote AWS Agent.")
@click.pass_obj
@click.option(
    "--assumable-role",
    help="ARN of AWS assumable role.",
    required=True,
)
@click.option(
    "--lambda-arn",
    help="ARN of AWS Lambda function.",
    required=True,
)
@click.option(
    "--external-id",
    help="AWS External ID.",
    required=False,
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_aws_agent(ctx, lambda_arn, **kwargs):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_aws_agent",
    ).create_agent(
        agent_type=REMOTE_AGENT,
        platform=AWS,
        storage=S3,
        auth_type=AWS_ASSUMABLE_ROLE,
        endpoint=lambda_arn,
        **kwargs,
    )


@agents.command(help="Register a Remote AWS Proxied Agent.")
@click.pass_obj
@click.option(
    "--authentication",
    help="Authentication type.",
    required=False,
    type=click.Choice(
        [OAUTH2_CLIENT_CREDENTIALS.lower(), OAUTH2_PASSWORD.lower()], case_sensitive=False
    ),
    default=OAUTH2_CLIENT_CREDENTIALS.lower(),
)
@click.option(
    "--proxy-endpoint",
    help="HTTP/HTTPS Proxy URL.",
    required=True,
)
@click.option(
    "--auth-url",
    help="OAuth2 Client Credentials - Authentication URL, used to get the access token.",
    required=False,
    cls=AdvancedOptions,
)
@click.option(
    "--client-id",
    help="OAuth2 Client Credentials - Client ID.",
    required=False,
    cls=AdvancedOptions,
    required_with_options=["client_secret", "auth_url"],
    mutually_exclusive_options=["username", "password"],
)
@click.option(
    "--client-secret",
    help=f"OAuth2 Client Credentials - Client Secret. {PASSWORD_VERBIAGE}",
    required=False,
    cls=AdvancedOptions,
    prompt_if_requested=True,
    required_with_options=["client_id", "auth_url"],
    mutually_exclusive_options=["username", "password"],
)
@click.option(
    "--username",
    help="OAuth2 Username/Password - Username.",
    required=False,
    cls=AdvancedOptions,
    required_with_options=["password", "auth_url", "authentication"],
    mutually_exclusive_options=["client_id", "client_secret"],
)
@click.option(
    "--password",
    help=f"OAuth2 Username/Password - Password. {PASSWORD_VERBIAGE}",
    required=False,
    cls=AdvancedOptions,
    prompt_if_requested=True,
    required_with_options=["username", "auth_url", "authentication"],
    mutually_exclusive_options=["client_id", "client_secret"],
)
@click.option(
    "--deployment-type",
    help="Specify if this AWS Agent is deployed as 'lambda' or 'generic' (e.g. ECS, EKS, EC2)",
    required=False,
    type=click.Choice(["lambda", "generic"], case_sensitive=False),
    default="lambda",
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(MUTUAL_TLS_OPTIONS)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_aws_proxied_agent(
    ctx,
    authentication: str,
    proxy_endpoint: str,
    auth_url: Optional[str] = None,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    deployment_type: Optional[str] = None,
    **kwargs,
):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_aws_proxied_agent",
    ).create_agent(
        agent_type=REMOTE_AGENT,
        platform=AWS_GENERIC if deployment_type == "generic" else AWS,
        storage=S3,
        auth_type=authentication.upper(),
        endpoint=proxy_endpoint,
        auth_url=auth_url,
        client_id=client_id,
        client_secret=client_secret,
        username=username,
        password=password,
        **kwargs,
    )


@agents.command(help="Register a Data Store Agent with Google Cloud Storage.")
@click.pass_obj
@click.option(
    "--key-file",
    help="JSON Key file if auth type is GCP JSON service account key.",
    type=click.Path(exists=True),
    required=True,
)
@click.option(
    "--bucket-name",
    help="Name of GCS bucket for data store.",
    required=True,
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_gcs_store(ctx, bucket_name, **kwargs):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_gcs_store",
    ).create_agent(
        agent_type=DATA_STORE_AGENT,
        platform=GCP,
        storage=GCS,
        auth_type=GCP_JSON_SERVICE_ACCOUNT_KEY,
        endpoint=bucket_name,
        **kwargs,
    )


@agents.command(help="Register a Remote GCP Agent.")
@click.pass_obj
@click.option(
    "--key-file",
    help="JSON Key file if auth type is GCP JSON service account key.",
    required=True,
)
@click.option(
    "--url",
    help="URL for accessing agent.",
    required=True,
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(MUTUAL_TLS_OPTIONS)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_gcp_agent(ctx, url, **kwargs):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_gcp_agent",
    ).create_agent(
        agent_type=REMOTE_AGENT,
        platform=GCP,
        storage=GCS,
        auth_type=GCP_JSON_SERVICE_ACCOUNT_KEY,
        endpoint=url,
        **kwargs,
    )


@agents.command(help="Register a Remote Azure Agent.")
@click.pass_obj
@click.option(
    "--app-key",
    help=f"App key from the Azure Function to use for authentication. {PASSWORD_VERBIAGE}",
    required=True,
    cls=AdvancedOptions,
    prompt_if_requested=True,
)
@click.option(
    "--url",
    help="URL for accessing agent.",
    required=True,
)
@add_common_options(DRY_RUN_OPTIONS)
@add_common_options(DC_ID_OPTION)
@add_common_options(MUTUAL_TLS_OPTIONS)
@add_common_options(UPDATE_AGENT_OPTIONS)
def register_azure_agent(ctx, url, **kwargs):
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents register_azure_agent",
    ).create_agent(
        agent_type=REMOTE_AGENT,
        platform=AZURE,
        storage=AZURE_BLOB,
        auth_type=AZURE_FUNCTION_APP_KEY,
        endpoint=url,
        **kwargs,
    )


@agents.command(help="Deregister an Agent.")
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent to deregister.", required=True)
def deregister(ctx, agent_id):
    """
    Deregister an Agent (deletes AgentModel from monolith)
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents deregister",
    ).delete_agent(agent_id)


@agents.command(help="Perform a health check of the Agent.")
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent.", required=True)
def health(ctx, agent_id):
    """
    Check the health of an Agent
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents health",
    ).check_agent_health(agent_id)


@agents.command(help="Upgrades the running version of an Agent.")
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent to upgrade.", required=True)
@click.option(
    "--image-tag",
    help="Image version to upgrade to.",
    required=False,
)
@click.option(
    "--params",
    "parameters",
    required=False,
    default=None,
    callback=validate_json_callback,
    help="""
              Parameters key,value pairs as JSON. If a key is not specified
              the existing (or default) value is used.
              \b
              \n
              E.g. --params '{"MemorySize":"1024", "ConcurrentExecutions": "25"}'
              """,
)  # \b disables wrapping
def upgrade(ctx, **kwargs):
    """
    Performs an upgrade of an Agent
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents upgrade",
    ).upgrade_agent(**kwargs)


@agents.command(help="List all agents in account.", name="list")
@click.pass_obj
@click.option(
    "--show-inactive",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Only list active agents.",
)
def list_agents(ctx, **kwargs):
    """
    List all Agents in account
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents list_agents",
    ).echo_agents(**kwargs)


@agents.command(help="Checks if telnet connection is usable from the agent.")
@click.pass_obj
@add_common_options(NETWORK_TEST_OPTIONS)
def test_telnet(ctx, **kwargs):
    """
    Network debugging utility to test telnet via the collector
    """
    CollectorNetworkTestService(
        config=ctx["config"],
        command_name="agents test_telnet",
    ).echo_telnet_test(**kwargs)


@agents.command(
    help="Tests if a destination exists and accepts requests. "
    "Opens a TCP Socket to a specific port from the agent."
)
@click.pass_obj
@add_common_options(NETWORK_TEST_OPTIONS)
def test_tcp_open(ctx, **kwargs):
    """
    Network debugging utility to test TCP open via the collector
    """
    CollectorNetworkTestService(
        config=ctx["config"],
        command_name="agents test_tcp_open",
    ).echo_tcp_open_test(**kwargs)


@agents.command(name="get-operation-logs", help="Returns the operation logs for a remote agent.")
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent.", required=True)
@click.option(
    "--start-time",
    help="Optional start time, for example: 2023-12-02T13:40:25Z. Defaults to 10 minutes ago.",
    required=False,
)
@click.option(
    "--end-time",
    help="Optional end time, for example: 2023-12-02T13:45:25Z. Defaults to now.",
    required=False,
)
@click.option(
    "--limit",
    help="Maximum number of log events to return, defaults to 1,000",
    type=click.INT,
    required=False,
)
@click.option(
    "--connection-type",
    help="Optional connection type to filter logs, for example snowflake, redshift, etc.",
    required=False,
)
def get_operation_logs(ctx, **kwargs):
    """
    Displays the operation logs for a remote Agent.
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents get_operation_logs",
    ).echo_operation_logs(**kwargs)


@agents.command(
    name="get-aws-upgrade-logs", help="Returns the upgrade logs for a remote AWS Agent."
)
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent.", required=True)
@click.option(
    "--limit",
    help="Maximum number of log events to return, defaults to 100",
    type=click.INT,
    required=False,
)
@click.option(
    "--start-time",
    help="Optional start time, for example: 2023-12-02T13:40:25Z. Defaults to 12 hours ago.",
    required=False,
)
def get_aws_upgrade_logs(ctx, **kwargs):
    """
    Displays the upgrade logs for a remote AWS Agent.
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents get_aws_upgrade_logs",
    ).echo_aws_upgrade_logs(**kwargs)


@agents.command(
    name="get-aws-template",
    help="Displays the current CloudFormation template in use by an AWS Agent, in YAML format.",
)
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent.", required=True)
def get_aws_template(ctx, **kwargs):
    """
    Displays the current CloudFormation template in use by an AWS Agent, in YAML format.
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents get_aws_template",
    ).echo_aws_template(**kwargs)


@agents.command(
    name="get-aws-template-parameters",
    help=(
        "Displays the current CloudFormation template parameters in use by an AWS Agent."
        " For Terraform it displays the current value for MemorySize and ConcurrentExecutions."
    ),
)
@click.pass_obj
@click.option("--agent-id", help="UUID of Agent.", required=True)
def get_aws_template_parameters(ctx, **kwargs):
    """
    Displays the current CloudFormation template parameters in use by an AWS Agent.
    """
    AgentService(
        config=ctx["config"],
        mc_client=create_mc_client(ctx),
        command_name="agents get_aws_template_parameters",
    ).echo_aws_template_parameters(**kwargs)



import click
import click_config_file

from montecarlodata import settings
from montecarlodata.common import create_mc_client
from montecarlodata.iac.mc_config_service import MonteCarloConfigService
from montecarlodata.monitors.monitor_service import MonitorService


@click.group(help="Manage monitors.")
def monitors():
    """
    Group for any monitor related subcommands
    """
    pass


@monitors.command(help="Compile monitor configuration.")
@click.option(
    "--project-dir",
    required=False,
    type=click.Path(exists=True),
    help="Base directory of MC project (where montecarlo.yml is located). "
    "By default, this is set to the current working directory",
)
@click.option(
    "--namespace",
    required=False,
    help="Namespace of monitors configuration. This value will be ignored if we find a"
    " namespace in montecarlo.yml",
)
@click.option(
    "--dbt-manifest",
    required=False,
    type=click.Path(exists=True),
    help="Path to dbt manifest used for resolving dbt ref() in monitor config",
)
@click.pass_obj
def compile(ctx, project_dir, namespace, dbt_manifest):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors compile",
        project_dir=project_dir,
        dbt_manifest_path=dbt_manifest,
    ).compile(namespace)


@monitors.command(help="Compile and apply monitor configuration.")
@click.option(
    "--project-dir",
    required=False,
    type=click.Path(exists=True),
    help="Base directory of MC project (where montecarlo.yml is located). "
    "By default, this is set to the current working directory",
)
@click.option(
    "--namespace",
    required=False,
    help="Namespace of monitors configuration. This value will be ignored if we find a"
    " namespace in montecarlo.yml",
)
@click.option(
    "--dry-run",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Dry run (just shows planned changes but doesn't apply them.)",
)
@click.option(
    "--dbt-manifest",
    required=False,
    type=click.Path(exists=True),
    help="Path to dbt manifest used for resolving dbt ref() in monitor config",
)
@click.option(
    "--auto-yes",
    is_flag=True,
    help="Skip any interactive approval.",
    default=False,
    show_default=True,
)
@click.option(
    "--create-non-ingested-tables",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="force create non-ingested tables if they don't exist.",
)
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def apply(
    ctx,
    project_dir,
    namespace,
    dry_run,
    dbt_manifest,
    auto_yes,
    create_non_ingested_tables,
):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors apply",
        project_dir=project_dir,
        dbt_manifest_path=dbt_manifest,
    ).apply(
        namespace,
        dry_run=dry_run,
        skip_confirmation=auto_yes,
        create_non_ingested_tables=create_non_ingested_tables,
    )


@monitors.command(help="Delete monitor configuration.")
@click.option(
    "--project-dir",
    required=False,
    type=click.Path(exists=True),
    help="Base directory of MC project (where montecarlo.yml is located). "
    "By default, this is set to the current working directory",
)
@click.option("--namespace", required=True, help="Namespace of monitors configuration.")
@click.option(
    "--dry-run",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Dry run (just shows planned changes but doesn't apply them.",
)
@click.pass_obj
@click_config_file.configuration_option(settings.OPTION_FILE_FLAG)
def delete(ctx, project_dir, namespace, dry_run):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors delete",
        project_dir=project_dir,
    ).delete(namespace, dry_run=dry_run)


@monitors.command(name="list", help="List monitors ordered by update recency.")
@click.option(
    "--limit",
    required=False,
    default=100,
    type=click.INT,
    show_default=True,
    help="Max number of monitors to list.",
)
@click.option(
    "--monitor-type",
    required=False,
    type=click.Choice(MonitorService.MONITOR_TYPES, case_sensitive=False),
    help="List monitors with monitor_type",
)
@click.option("--namespace", required=False, help="List only monitors in this namespace")
@click.pass_obj
def list_monitors(ctx, monitor_type, limit, namespace):
    monitor_types = None
    namespaces = [namespace] if namespace else None
    if monitor_type:
        if monitor_type == MonitorService.PSEUDO_MONITOR_TYPE_CB_COMPATIBLE:
            monitor_type = MonitorService.MONITOR_TYPE_CUSTOM_SQL
        monitor_types = [monitor_type]
    MonitorService(
        client=create_mc_client(ctx),
        command_name="monitors list_monitors",
    ).list_monitors(limit, namespaces, monitor_types)


@monitors.command(name="namespaces", help="List all namespaces.")
@click.option(
    "--limit",
    required=False,
    default=100,
    type=click.INT,
    show_default=True,
    help="Max number of namespaces to list.",
)
@click.pass_obj
def list_namespaces(ctx, limit):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors list_namespaces",
    ).list_namespaces(limit)


@monitors.command(help="Generate MC monitor config YAML from dbt tests.")
@click.option(
    "--output-path",
    required=True,
    type=click.Path(),
    help="Path for MC monitor config output YAML",
)
@click.option(
    "--dbt-manifest",
    required=True,
    type=click.Path(exists=True),
    help="Path to dbt manifest containing tests to generate MC monitor config for",
)
@click.option(
    "--test-type",
    required=False,
    multiple=True,
    help="Filter for types of dbt tests to convert. Can pass multiple. Defaults to all."
    " Pass SINGULAR to filter to singular tests without a shared type.",
)
@click.option(
    "--label",
    required=False,
    multiple=True,
    help="Label to apply to all generated MC monitors. Can pass multiple.",
)
@click.pass_obj
def generate_from_dbt_tests(ctx, dbt_manifest, output_path, test_type, label):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors generate_from_dbt_tests",
        dbt_manifest_path=dbt_manifest,
    ).generate_from_dbt_tests(output_path, test_type, label)


@monitors.command(
    help="Convert monitors from UI to monitors configuration, "
    "exporting the monitors as YAML in the process."
)
@click.option(
    "--namespace",
    required=True,
    type=click.STRING,
    help="Namespace for the exported monitors.",
)
@click.option(
    "--project-dir",
    required=True,
    type=click.Path(),
    help="Path to a directory to export a MC monitor config project to.",
)
@click.option(
    "--monitors-file",
    required=False,
    type=click.STRING,
    help="File with monitor UUIDs to convert. One line per monitor.",
)
@click.option(
    "--dry-run",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Dry run (only export the monitors as monitors as code, "
    "but do not convert existing ones).",
)
@click.pass_obj
def convert_to_mac(ctx, namespace, dry_run, project_dir, monitors_file):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors convert_to_mac",
    ).convert_to_mac(
        namespace,
        project_dir,
        monitors_file,
        all_monitors=not monitors_file,
        dry_run=dry_run,
    )


@monitors.command(help="Convert monitors from a monitors as code namespace to UI monitors.")
@click.option(
    "--namespace",
    required=True,
    type=click.STRING,
    help="Namespace to convert.",
)
@click.option(
    "--dry-run",
    required=False,
    default=False,
    show_default=True,
    is_flag=True,
    help="Dry run (just shows planned changes but doesn't apply them.)",
)
@click.pass_obj
def convert_to_ui(ctx, namespace, dry_run):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors convert_to_ui",
    ).convert_to_ui(namespace, dry_run=dry_run)


@monitors.command(
    help="Export monitors in a monitor as code namespace upgraded to the latest "
    "version of monitors of monitors as code. Will only export monitors that "
    "can be upgraded."
)
@click.option(
    "--namespace",
    required=True,
    type=click.STRING,
    help="Namespace to export as latest.",
)
@click.option(
    "--monitors-file",
    required=False,
    type=click.STRING,
    help="File with monitor UUIDs to export as latest. One line per monitor.",
)
@click.pass_obj
def export_as_latest(ctx, namespace, monitors_file):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors export_as_latest",
    ).export_as_latest(namespace, monitors_file=monitors_file)


@monitors.command(help="Get the monitors configuration for a given namespace.")
@click.option(
    "--namespace",
    required=True,
    type=click.STRING,
    help="Namespace for the exported monitors.",
)
@click.option(
    "--project-dir",
    required=True,
    type=click.Path(),
    help="Path to a directory to export a the template to.",
)
@click.pass_obj
def get_template(ctx, namespace, project_dir):
    MonteCarloConfigService(
        config=ctx["config"],
        pycarlo_client=create_mc_client(ctx),
        command_name="monitors get_template",
    ).get_template(namespace, project_dir)



import json
import os
import pathlib
import shutil
from unittest import TestCase
from unittest.mock import ANY, Mock, call, patch

import click
import responses
from box import Box, BoxList
from click.testing import CliRunner
from freezegun import freeze_time
from parameterized import parameterized
from pycarlo.core import Client

from montecarlodata.common.data import MonolithResponse
from montecarlodata.iac.commands import (
    compile as compile_config,
)
from montecarlodata.iac.commands import (
    convert_to_mac,
    convert_to_ui,
    export_as_latest,
    get_template,
)
from montecarlodata.iac.mc_config_service import MonteCarloConfigService
from montecarlodata.queries.iac import (
    CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
    GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
)
from montecarlodata.utils import GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG


class MonteCarloConfigServiceTest(TestCase):
    maxDiff = None

    @patch("os.getcwd")
    def setUp(self, getcwd) -> None:
        self._request_wrapper_mock = Mock(spec=GqlWrapper)
        self._pycarlo_client = Mock(spec=Client)
        self._print_func = Mock(spec=click.echo)
        self.project_dir = self._get_project_dir("standalone_configs")
        getcwd.return_value = self.project_dir
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            pycarlo_client=self._pycarlo_client,
            print_func=self._print_func,
        )

    def test_standalone_configs(self):
        namespace = "test"
        files, template, _ = self.service.compile(namespace)
        self.assertEqual(len(files), 4)
        self.assertEqual(
            json.loads(json.dumps(template)),
            self._get_standalone_config_compiled_template(namespace),
        )

    def test_namespace_override(self):
        project_dir = self._get_project_dir("namespace_override")
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=project_dir,
            pycarlo_client=self._pycarlo_client,
        )
        files, template, _ = self.service.compile("test")

        self.assertEqual(len(files), 3)
        self.assertEqual(
            json.loads(json.dumps(template)),
            {
                "custom_namespace": {
                    "field_health": [
                        {
                            "table": "analytics:prod_lineage.lineage_nodes",
                            "timestamp_field": "created",
                        },
                        {"table": "analytics:prod.abc"},
                    ],
                    "freshness": [
                        {
                            "freshness_threshold": 30,
                            "schedule": {
                                "interval_minutes": 30,
                                "start_time": "2021-07-27T19:51:00",
                                "type": "fixed",
                            },
                            "table": "analytics:prod.abc",
                        }
                    ],
                },
                "infra": {
                    "dimension_tracking": [
                        {
                            "field": "account_id",
                            "table": "analytics:prod.customer_360",
                            "yaml_file_name": f"{project_dir}/monitors.yaml",
                            "yaml_line": 5,
                        }
                    ]
                },
                "test": {
                    "dimension_tracking": [
                        {
                            "field": "account_id",
                            "table": "analytics:prod.customer_360",
                            "yaml_file_name": f"{project_dir}/monitors_other.yml",
                            "yaml_line": 3,
                        }
                    ]
                },
            },
        )

    @parameterized.expand(
        [
            ("namespace_override", "montecarlo.yml"),
            ("namespace_override_yaml", "montecarlo.yaml"),
        ]
    )
    def test_namespace_override_no_default_namesapce(self, project_dir_name, config_file_name):
        project_dir = self._get_project_dir(project_dir_name)
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=project_dir,
            pycarlo_client=self._pycarlo_client,
        )
        files, template, errors = self.service.compile(None, abort_on_error=False)

        self.assertEqual(len(files), 3)
        self.assertEqual(
            json.loads(json.dumps(template)),
            {
                "custom_namespace": {
                    "field_health": [
                        {
                            "table": "analytics:prod_lineage.lineage_nodes",
                            "timestamp_field": "created",
                        },
                        {"table": "analytics:prod.abc"},
                    ],
                    "freshness": [
                        {
                            "freshness_threshold": 30,
                            "schedule": {
                                "interval_minutes": 30,
                                "start_time": "2021-07-27T19:51:00",
                                "type": "fixed",
                            },
                            "table": "analytics:prod.abc",
                        }
                    ],
                },
                "infra": {
                    "dimension_tracking": [
                        {
                            "field": "account_id",
                            "table": "analytics:prod.customer_360",
                            "yaml_file_name": f"{project_dir}/monitors.yaml",
                            "yaml_line": 5,
                        }
                    ]
                },
            },
        )
        self.assertEqual(len(errors), 1)
        file_name = next(iter(errors))
        self.assertTrue(
            f"tests/iac/test_resources/{project_dir_name}/monitors_other.yml" in file_name
        )
        self.assertTrue(
            "A default namespace need to be passed through command line --namespace "
            f"or set default namespace in {config_file_name} or an override namespace needs to be set in file"
            in errors[file_name][0]
        )

    def test_namespace_override_default_namespace_montecarlo_yaml(self):
        project_dir = self._get_project_dir("namespace_override_montecarlo")
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=project_dir,
            pycarlo_client=self._pycarlo_client,
        )

        files, template, errors = self.service.compile(None, abort_on_error=False)

        self.assertEqual(len(files), 3)
        self.assertEqual(
            json.loads(json.dumps(template)),
            {
                "custom_namespace": {
                    "field_health": [
                        {
                            "table": "analytics:prod_lineage.lineage_nodes",
                            "timestamp_field": "created",
                        },
                        {"table": "analytics:prod.abc"},
                    ],
                    "freshness": [
                        {
                            "freshness_threshold": 30,
                            "schedule": {
                                "interval_minutes": 30,
                                "start_time": "2021-07-27T19:51:00",
                                "type": "fixed",
                            },
                            "table": "analytics:prod.abc",
                        }
                    ],
                },
                "infra": {
                    "dimension_tracking": [
                        {
                            "field": "account_id",
                            "table": "analytics:prod.customer_360",
                            "yaml_file_name": f"{project_dir}/monitors.yml",
                            "yaml_line": 5,
                        }
                    ]
                },
                "test_montecarlo": {
                    "dimension_tracking": [
                        {
                            "field": "account_id",
                            "table": "analytics:prod.customer_360",
                            "yaml_file_name": f"{project_dir}/monitors_other.yml",
                            "yaml_line": 3,
                        }
                    ]
                },
            },
        )
        self.assertEqual(len(errors), 0)

    def test_embedded_dbt_configs(self):
        project_dir = self._get_project_dir("embedded_dbt_configs")
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=project_dir,
            pycarlo_client=self._pycarlo_client,
        )
        files, template, _ = self.service.compile("test")

        self.assertEqual(len(files), 4)
        self.assertEqual(
            json.loads(json.dumps(template)),
            {
                "test": {
                    "field_health": [
                        {
                            "table": "analytics:prod_lineage.lineage_nodes",
                            "timestamp_field": "created",
                        },
                        {"table": "analytics:prod.abc"},
                        {
                            "table": "analytics:prod.client_hub",
                            "fields": ["account_id"],
                            "yaml_line": 3,
                            "yaml_file_name": f"{project_dir}/dir1/dir2/monitors.yml",
                        },
                    ],
                    "freshness": [
                        {
                            "table": "analytics:prod.abc",
                            "freshness_threshold": 30,
                            "schedule": {
                                "type": "fixed",
                                "interval_minutes": 30,
                                "start_time": "2021-07-27T19:51:00",
                            },
                        }
                    ],
                    "dimension_tracking": [
                        {
                            "table": "analytics:prod.customer_360",
                            "field": "account_id",
                            "yaml_line": 3,
                            "yaml_file_name": f"{project_dir}/dir1/monitors.yml",
                        }
                    ],
                },
            },
        )

    def test_embedded_dbt_configs_with_refs(self):
        project_dir = self._get_project_dir("embedded_dbt_configs_with_refs")
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=project_dir,
            pycarlo_client=self._pycarlo_client,
            dbt_manifest_path=f"{project_dir}/manifest.json",
        )
        files, template, _ = self.service.compile("test")

        self.assertEqual(len(files), 4)
        self.assertEqual(
            json.loads(json.dumps(template)),
            {
                "test": {
                    "custom_sql": [
                        {
                            "notes": "{{query_result:link, connection_type}}",
                            "sampling_sql": "select * from analytics.prod.customer_360",
                            "sql": "select * from analytics.prod.def a join "
                            "analytics.prod.client_hub b on a.id = b.id",
                            "variables": {"foo": ["bar analytics.prod.def"]},
                        }
                    ],
                    "dimension_tracking": [
                        {
                            "field": "account_id",
                            "table": "analytics:prod.customer_360",
                            "yaml_file_name": f"{project_dir}/dir1/monitors.yml",
                            "yaml_line": 3,
                        }
                    ],
                    "field_health": [
                        {
                            "table": "analytics:prod_lineage.lineage_nodes",
                            "timestamp_field": "created",
                        },
                        {"table": "analytics:prod.def"},
                        {
                            "fields": ["account_id"],
                            "name": "a analytics:prod.client_hub or analytics:prod.def",
                            "table": "analytics:prod.client_hub",
                            "yaml_file_name": f"{project_dir}/dir1/dir2/monitors.yml",
                            "yaml_line": 3,
                        },
                    ],
                    "freshness": [
                        {
                            "freshness_threshold": 30,
                            "schedule": {
                                "interval_minutes": 30,
                                "start_time": "2021-07-27T19:51:00",
                                "type": "fixed",
                            },
                            "tables": [
                                "analytics:prod.def",
                                "analytics:prod.client_hub",
                            ],
                        }
                    ],
                }
            },
        )

    def test_invalid_configs(self):
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=self._get_project_dir("invalid_configs"),
            pycarlo_client=self._pycarlo_client,
        )
        self.service._abort_on_error = False

        files, template, errors_by_file = self.service.compile("test", abort_on_error=False)
        errors = sorted(list(errors_by_file.items()), key=lambda x: x[0])

        self.assertEqual(len(errors), 3)
        file, error = errors[0]
        self.assertTrue(file.endswith("dir1/dir2/monitors.yml"))
        self.assertEqual(error, ['"custom_sql" property should be a list.'])
        file, error = errors[1]
        self.assertTrue(file.endswith("dir1/monitors.yml"))
        self.assertEqual(error, ['"field_health" property should be a list.'])
        file, error = errors[2]
        self.assertTrue(file.endswith("dir1/repeated_key.yml"))
        self.assertEqual(
            error, ['Failed to parse YAML: Duplicate "dimension_tracking" key found in YAML.']
        )

    def test_yaml_references(self):
        project_dir = self._get_project_dir("yaml_references")
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            project_dir=self._get_project_dir("yaml_references"),
            pycarlo_client=self._pycarlo_client,
        )
        self.service._abort_on_error = False

        files, template, errors_by_file = self.service.compile("test", abort_on_error=False)
        self.assertEqual(
            template,
            {
                "test": {
                    "dimension_tracking": BoxList(
                        [
                            {
                                "lookback_days": 2,
                                "name": "dt-1-mac",
                                "yaml_line": 8,
                                "yaml_file_name": f"{project_dir}/dir1/repeated_key_references.yml",
                            },
                            {
                                "labels": ["Common (product)"],
                                "name": "dt-2-mac",
                                "yaml_line": 8,
                                "yaml_file_name": f"{project_dir}/dir1/yaml_references.yml",
                            },
                        ]
                    )
                }
            },
        )
        errors = sorted(list(errors_by_file.items()), key=lambda x: x[0])

        self.assertEqual(len(errors), 0)

    def test_apply(self):
        namespace = "foo"
        update_uuid = "eeeeeeee-58fd-44d0-8b2d-eeeeeeeeeeee"
        update_response = MonolithResponse(
            data={
                "response": {
                    "updateUuid": update_uuid,
                    "errorsAsJson": "{}",
                    "warningsAsJson": "{}",
                }
            }
        )
        poll_response = MonolithResponse(
            data={
                "state": "APPLIED",
                "resourceModifications": [
                    {
                        "type": "ResourceModificationType.UPDATE",
                        "description": "Monitor: type=stats, table=analytics:prod.client_hub",
                        "resourceAsJson": json.dumps(
                            {
                                "uuid": "ed4d07c3-58fd-44d0-8b2d-c1b020f45a69",
                                "resource": None,
                                "name": (
                                    "monitor|type=stats|table=analytics:prod.client_hub|"
                                    "timestamp_field=<<NULL>>|where_condition=<<NULL>>"
                                ),
                                "table": "analytics:prod.customer_360",
                                "type": "stats",
                                "fields": [],
                                "timestamp_field": None,
                                "where_condition": None,
                            }
                        ),
                    },
                    {
                        "type": "ResourceModificationType.UPDATE",
                        "description": "Monitor: type=categories, table=analytics:prod.customer_360",
                        "resourceAsJson": json.dumps(
                            {
                                "uuid": "ec3b0a80-d088-4dbe-acf5-150caf041574",
                                "resource": None,
                                "name": (
                                    "monitor|type=categories|table=analytics:prod.customer_360|"
                                    "timestamp_field=<<NULL>>|where_condition=<<NULL>>|fields=account_id"
                                ),
                                "table": "analytics:prod.customer_360",
                                "type": "categories",
                                "fields": ["account_id"],
                                "timestamp_field": None,
                                "where_condition": None,
                            }
                        ),
                    },
                ],
                "changesApplied": True,
                "errorsAsJson": "{}",
                "warningsAsJson": "{}",
            }
        )
        self._request_wrapper_mock.make_request_v2.side_effect = [
            update_response,
            poll_response,
            update_response,
            poll_response,
        ]

        responses = self.service.apply(namespace, skip_confirmation=True)

        compiled_template = self._get_standalone_config_compiled_template(namespace)

        self.assertEqual(
            [
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=json.dumps(compiled_template[namespace]),
                        dryRun=True,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=json.dumps(compiled_template[namespace]),
                        dryRun=False,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
            ],
            self._request_wrapper_mock.make_request_v2.mock_calls,
        )

        self.assertEqual(responses[0].errors, {})
        self.assertEqual(len(responses[0].resource_modifications), 2)

    def test_apply_dry_run(self):
        namespace = "foo"
        update_uuid = "eeeeeeee-58fd-44d0-8b2d-eeeeeeeeeeee"
        update_response = MonolithResponse(
            data={
                "response": {
                    "updateUuid": update_uuid,
                    "errorsAsJson": "{}",
                    "warningsAsJson": "{}",
                }
            }
        )
        poll_response = MonolithResponse(
            data={
                "state": "APPLIED",
                "resourceModifications": [
                    {
                        "type": "ResourceModificationType.UPDATE",
                        "description": "Monitor: type=stats, table=analytics:prod.client_hub",
                        "resourceAsJson": '{"uuid": "ed4d07c3-58fd-44d0-8b2d-c1b020f45a69", "resource": null, "name": "monitor|type=stats|table=analytics:prod.client_hub|timestamp_field=<<NULL>>|where_condition=<<NULL>>", "table": "analytics:prod.customer_360", "type": "stats", "fields": [], "timestamp_field": null, "where_condition": null}',
                    },
                    {
                        "type": "ResourceModificationType.UPDATE",
                        "description": "Monitor: type=categories, table=analytics:prod.customer_360",
                        "resourceAsJson": '{"uuid": "ec3b0a80-d088-4dbe-acf5-150caf041574", "resource": null, "name": "monitor|type=categories|table=analytics:prod.customer_360|timestamp_field=<<NULL>>|where_condition=<<NULL>>|fields=account_id", "table": "analytics:prod.customer_360", "type": "categories", "fields": ["account_id"], "timestamp_field": null, "where_condition": null}',
                    },
                ],
                "changesApplied": True,
                "errorsAsJson": "{}",
                "warningsAsJson": "{}",
            }
        )
        self._request_wrapper_mock.make_request_v2.side_effect = [
            update_response,
            poll_response,
        ]

        responses = self.service.apply(namespace, dry_run=True)

        compiled_template = self._get_standalone_config_compiled_template(namespace)

        self._request_wrapper_mock.make_request_v2.assert_has_calls(
            [
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=json.dumps(compiled_template[namespace]),
                        dryRun=True,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
            ]
        )

        self.assertEqual(responses[0].errors, {})
        self.assertEqual(len(responses[0].resource_modifications), 2)

    def test_apply_with_errors(self):
        namespace = "foo"
        update_uuid = "eeeeeeee-58fd-44d0-8b2d-eeeeeeeeeeee"
        self._request_wrapper_mock.make_request_v2.return_value = MonolithResponse(
            data={
                "response": {
                    "updateUuid": update_uuid,
                    "errorsAsJson": '{"validation_errors": {"monitors": {"0": {"type": ["Unknown field."]}}}}',
                    "warningsAsJson": "{}",
                }
            }
        )

        responses = self.service.apply(namespace, abort_on_error=False)
        compiled_template = self._get_standalone_config_compiled_template(namespace)

        self._request_wrapper_mock.make_request_v2.assert_called_once_with(
            query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
            operation="createOrUpdateMonteCarloConfigTemplateAsync",
            service="iac_service",
            variables=dict(
                namespace=namespace,
                configTemplateJson=json.dumps(compiled_template[namespace]),
                dryRun=True,
                misconfiguredAsWarning=True,
                resource=None,
                createNonIngestedTables=False,
            ),
        )

        self.assertEqual(
            responses[0].errors,
            {"validation_errors": {"monitors": {"0": {"type": ["Unknown field."]}}}},
        )
        self.assertEqual(len(responses[0].resource_modifications), 0)

    @parameterized.expand(
        [
            (
                "Failed to apply changes.",
                {"system_error": "Problem applying template resource: ..."},
                "FAILED",
                '{"system_error": "Problem applying template resource: ..."}',
            ),
            ("Update skipped in favor of more recent update.", {}, "SKIPPED", "{}"),
        ]
    )
    def test_apply_with_resource_apply_errors(
        self,
        expected_msg: str,
        expected_errors: dict,
        result_state: str,
        result_errors: str,
    ):
        namespace = "foo"
        update_uuid = "eeeeeeee-58fd-44d0-8b2d-eeeeeeeeeeee"
        update_response = MonolithResponse(
            data={
                "response": {
                    "updateUuid": update_uuid,
                    "errorsAsJson": "{}",
                }
            }
        )
        dry_run_poll_response = MonolithResponse(
            data={
                "state": "APPLIED",
                "resourceModifications": [
                    {
                        "type": "ResourceModificationType.UPDATE",
                        "description": "Monitor: type=stats, table=analytics:prod.client_hub",
                        "resourceAsJson": '{"uuid": "ed4d07c3-58fd-44d0-8b2d-c1b020f45a69", "resource": null, "name": "monitor|type=stats|table=analytics:prod.client_hub|timestamp_field=<<NULL>>|where_condition=<<NULL>>", "table": "analytics:prod.customer_360", "type": "stats", "fields": [], "timestamp_field": null, "where_condition": null}',
                    },
                    {
                        "type": "ResourceModificationType.UPDATE",
                        "description": "Monitor: type=categories, table=analytics:prod.customer_360",
                        "resourceAsJson": '{"uuid": "ec3b0a80-d088-4dbe-acf5-150caf041574", "resource": null, "name": "monitor|type=categories|table=analytics:prod.customer_360|timestamp_field=<<NULL>>|where_condition=<<NULL>>|fields=account_id", "table": "analytics:prod.customer_360", "type": "categories", "fields": ["account_id"], "timestamp_field": null, "where_condition": null}',
                    },
                ],
                "changesApplied": False,
                "errorsAsJson": "{}",
                "warningsAsJson": "{}",
            }
        )
        wet_poll_response = MonolithResponse(
            data={
                "state": result_state,
                "resourceModifications": [],
                "changesApplied": False,
                "errorsAsJson": result_errors,
            }
        )
        self._request_wrapper_mock.make_request_v2.side_effect = [
            update_response,
            dry_run_poll_response,
            update_response,
            wet_poll_response,
        ]

        responses = self.service.apply(
            namespace,
            skip_confirmation=True,
            abort_on_error=False,
        )

        compiled_template = self._get_standalone_config_compiled_template(namespace)

        self._request_wrapper_mock.make_request_v2.assert_has_calls(
            [
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=json.dumps(compiled_template[namespace]),
                        dryRun=True,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=json.dumps(compiled_template[namespace]),
                        dryRun=False,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
            ]
        )

        self.assertEqual(responses[0].errors, expected_errors)
        self.assertEqual(len(responses[0].resource_modifications), 0)
        self._print_func.assert_has_calls([call(expected_msg)])

    def test_misconfigured_warnings(self):
        namespace = "foo"
        update_uuid = "eeeeeeee-58fd-44d0-8b2d-eeeeeeeeeeee"
        update_response = MonolithResponse(
            data={
                "response": {
                    "updateUuid": update_uuid,
                    "errorsAsJson": "{}",
                    "warningsAsJson": "{}",
                }
            }
        )
        poll_response = MonolithResponse(
            data={
                "state": "APPLIED",
                "resourceModifications": [
                    {
                        "type": "ResourceModificationType.CREATE",
                        "description": (
                            "Freshness SLO: "
                            "tables=['analytics:prod.client_hub', 'analytics:prod.client_warehouses'] "
                            "freshness_threshold=1"
                        ),
                        "resourceAsJson": json.dumps(
                            {
                                "uuid": None,
                                "resource": None,
                                "name": "freshnessrule|tables=(multiple)",
                                "description": None,
                                "notes": None,
                                "labels": [],
                                "schedule": {
                                    "type": "fixed",
                                    "interval_minutes": 1,
                                    "interval_crontab": None,
                                    "start_time": "2021-07-27T19:00:00",
                                },
                                "table": None,
                                "tables": [
                                    "analytics:prod.client_hub",
                                    "analytics:prod.client_warehouses",
                                ],
                                "freshness_threshold": 1,
                            }
                        ),
                    }
                ],
                "changesApplied": False,
                "errorsAsJson": "{}",
                "warningsAsJson": json.dumps(
                    {
                        "misconfigured_warnings": [
                            {
                                "title": "High breaching Freshness SLOs are defined, adjust freshness_threshold in",
                                "items": [
                                    (
                                        "analytics:prod.client_hub: "
                                        "freshness_threshold is 1, the expected minimum threshold is 464"
                                    ),
                                    (
                                        "analytics:prod.client_warehouses: "
                                        "freshness_threshold is 1, the expected minimum threshold is 1440"
                                    ),
                                ],
                            }
                        ]
                    }
                ),
            }
        )
        self._request_wrapper_mock.make_request_v2.side_effect = [
            update_response,
            poll_response,
            update_response,
            poll_response,
        ]

        responses = self.service.apply(namespace, abort_on_error=False, skip_confirmation=True)

        self._request_wrapper_mock.make_request_v2.assert_has_calls(
            [
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=ANY,
                        dryRun=True,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
                call(
                    query=CREATE_OR_UPDATE_MONTE_CARLO_CONFIG_TEMPLATE_ASYNC,
                    operation="createOrUpdateMonteCarloConfigTemplateAsync",
                    service="iac_service",
                    variables=dict(
                        namespace=namespace,
                        configTemplateJson=ANY,
                        dryRun=False,
                        misconfiguredAsWarning=True,
                        resource=None,
                        createNonIngestedTables=False,
                    ),
                ),
                call(
                    query=GET_MONTE_CARLO_CONFIG_TEMPLATE_UPDATE_STATE,
                    operation="getMonteCarloConfigTemplateUpdateState",
                    service="iac_service",
                    variables=dict(updateUuid=update_uuid),
                ),
            ]
        )

        self.assertEqual(
            responses[0].warnings,
            {
                "misconfigured_warnings": [
                    {
                        "items": [
                            (
                                "analytics:prod.client_hub: freshness_threshold is 1, the expected "
                                "minimum threshold is 464"
                            ),
                            (
                                "analytics:prod.client_warehouses: freshness_threshold is 1, the expected "
                                "minimum threshold is 1440"
                            ),
                        ],
                        "title": "High breaching Freshness SLOs are defined, adjust freshness_threshold in",
                    }
                ]
            },
        )
        self.assertEqual(len(responses[0].resource_modifications), 1)
        self._print_func.assert_has_calls([call("Changes successfully applied.")])

    def test_list_namespaces(self):
        self._pycarlo_client.return_value = self._mc_config_templates_response(self.LIMIT)
        self.service.list_namespaces(self.LIMIT)
        self._print_func.assert_called_once_with(self.NAMESPACES_TABLE)

    def test_list_namespaces_with_more_available(self):
        self._pycarlo_client.return_value = self._mc_config_templates_response(self.LIMIT + 1)
        self.service.list_namespaces(self.LIMIT)
        expected_calls = [
            call(self.NAMESPACES_TABLE),
            call(self.service.MORE_NS_MESSAGE),
        ]
        self._print_func.assert_has_calls(expected_calls)

    def test_compile_command_no_namespace(self):
        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        dir = self._get_project_dir("namespace_override")
        result = runner.invoke(
            compile_config,
            obj={"config": ctx},
            args=["--project-dir", dir, "--namespace", None],
        )

        self.assertEqual(
            result.output,
            (
                "\n"
                "Gathering monitor configuration files.\n"
                " - "
                f"{dir}/dbt_models/schema.yml "
                "- Embedded monitor configuration found.\n"
                " - "
                f"{dir}/monitors.yaml "
                "- Monitor configuration found.\n"
                "\n"
                "Configuration validation errors:\n"
                " - File: "
                f"{dir}/monitors_other.yml\n"
                "    - A default namespace need to be passed through command line --namespace "
                "or set default namespace in montecarlo.yml or an override namespace needs to "
                "be set in file: "
                f"{dir}/monitors_other.yml\n"
                "\n"
                "Error - Errors encountered, exiting.\n"
                "Aborted!\n"
            ),
        )

    def test_compile_command_ns_config(self):
        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        dir = self._get_project_dir("namespace_override_montecarlo")
        result = runner.invoke(
            compile_config,
            obj={"config": ctx},
            args=["--project-dir", dir, "--namespace", None],
        )

        self.assertEqual(
            result.output,
            (
                "\n"
                "Gathering monitor configuration files.\n"
                "namespace: test_montecarlo found in montecarlo.yml ignoring value passed "
                "with --namespace=None\n"
                " - "
                f"{dir}/dbt_models/schema.yml "
                "- Embedded monitor configuration found.\n"
                " - "
                f"{dir}/monitors.yml "
                "- Monitor configuration found.\n"
                " - "
                f"{dir}/monitors_other.yml "
                "- Monitor configuration found.\n"
            ),
        )

    def test_compile_command_ns_param(self):
        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        dir = self._get_project_dir("namespace_override")
        result = runner.invoke(
            compile_config,
            obj={"config": ctx},
            args=["--project-dir", dir, "--namespace", "test-mont"],
        )

        self.assertEqual(
            result.output,
            (
                "\n"
                "Gathering monitor configuration files.\n"
                " - "
                f"{dir}/dbt_models/schema.yml "
                "- Embedded monitor configuration found.\n"
                " - "
                f"{dir}/monitors.yaml "
                "- Monitor configuration found.\n"
                " - "
                f"{dir}/monitors_other.yml "
                "- Monitor configuration found.\n"
            ),
        )

    def test_compile_command_ns_config_param(self):
        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        dir = self._get_project_dir("namespace_override_montecarlo")
        result = runner.invoke(
            compile_config,
            obj={"config": ctx},
            args=["--project-dir", dir, "--namespace", "test-namespace"],
        )

        self.assertEqual(
            result.output,
            (
                "\n"
                "Gathering monitor configuration files.\n"
                "namespace: test_montecarlo found in montecarlo.yml ignoring value passed "
                "with --namespace=test-namespace\n"
                " - "
                f"{dir}/dbt_models/schema.yml "
                "- Embedded monitor configuration found.\n"
                " - "
                f"{dir}/monitors.yml "
                "- Monitor configuration found.\n"
                " - "
                f"{dir}/monitors_other.yml "
                "- Monitor configuration found.\n"
            ),
        )

    @freeze_time("2020-01-01")
    def test_generate_from_dbt_tests(self):
        project_dir = self._get_project_dir("import_dbt_tests")
        output_path = f"{project_dir}/dbt.yml"
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            pycarlo_client=self._pycarlo_client,
            dbt_manifest_path=f"{project_dir}/manifest.json",
        )
        self.service.generate_from_dbt_tests(output_path)

        with open(f"{project_dir}/expected.yml", "r") as f:
            expected_yaml = f.read()

        with open(output_path, "r") as f:
            output_yaml = f.read()

        self.assertEqual(expected_yaml, output_yaml)

    @freeze_time("2020-01-01")
    def test_generate_from_dbt_tests_when_filtered_and_labeled(self):
        project_dir = self._get_project_dir("import_dbt_tests")
        output_path = f"{project_dir}/dbt.yml"
        self.service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            pycarlo_client=self._pycarlo_client,
            dbt_manifest_path=f"{project_dir}/manifest.json",
        )
        self.service.generate_from_dbt_tests(
            output_path,
            test_types=["SINGULAR", "unique"],
            labels=["foo", "bar"],
        )

        with open(output_path, "r") as f:
            output_yaml = f.read()

        self.assertEqual(
            """\
montecarlo:
  custom_sql:
  - name: test.analytics.assert_not_null_recent_metrics_resource_id
    description: test.analytics.assert_not_null_recent_metrics_resource_id
    sql: |
      SELECT
          *
      FROM
          personal.dbt_bneville.recent_metrics
      WHERE
          resource_id IS NULL AND
          metric != 'total_views'
    schedule:
      type: fixed
      start_time: '2020-01-01T00:00:00+00:00'
      interval_minutes: 720
    comparisons:
    - type: threshold
      operator: GT
      threshold_value: 0
    labels:
    - bar
    - foo
  - name: test.analytics.unique_table_field_metrics_multi_key.14e928c2e4
    description: test.analytics.unique_table_field_metrics_multi_key.14e928c2e4
    sql: |
      select
          multi_key as unique_field,
          count(*) as n_records

      from personal.dbt_bneville.table_field_metrics
      where multi_key is not null
      group by multi_key
      having count(*) > 1
    schedule:
      type: fixed
      start_time: '2020-01-01T00:00:00+00:00'
      interval_minutes: 720
    comparisons:
    - type: threshold
      operator: GT
      threshold_value: 0
    labels:
    - bar
    - foo\n""",
            output_yaml,
        )

    def _get_project_dir(self, dir_name: str):
        return os.path.join(pathlib.Path(__file__).parent.resolve(), "test_resources", dir_name)

    def _get_temp_dir(self, dir_name: str):
        return os.path.join(
            pathlib.Path(__file__).parent.resolve(), "test_resources", "temp", dir_name
        )

    LIMIT = 2
    NAMESPACES_TABLE = """\
╒═════════════╤══════════════════════════════════╕
│ Namespace   │ Last Update Time                 │
╞═════════════╪══════════════════════════════════╡
│ namespace_1 │ 2000-01-01 00:00:00.000000+00:00 │
├─────────────┼──────────────────────────────────┤
│ namespace_2 │ 2000-01-01 00:00:00.000000+00:00 │
╘═════════════╧══════════════════════════════════╛"""

    @responses.activate
    def test_convert_to_ui(self):
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "convertConfigTemplateToUiMonitors": {
                        "response": {
                            "monitors": [
                                {"uuid": "uuid1", "name": "test1"},
                                {"uuid": "uuid2", "name": "test2"},
                            ],
                        }
                    }
                }
            },
        )

        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        result = runner.invoke(
            convert_to_ui,
            obj={"config": ctx},
            args=["--namespace", "test-namespace", "--dry-run"],
        )
        self.assertEqual(
            result.output,
            """Exported the following monitors to UI:
* uuid1: test1
* uuid2: test2
""",
        )

    @patch("os.getcwd")
    @responses.activate
    def test_convert_to_mac(self, getcwd):
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "convertUiMonitorsToConfigTemplate": {
                        "response": {
                            "configTemplateAsDict": json.dumps({"foo": "bar"}),
                            "errors": [],
                        },
                    },
                }
            },
        )

        dir = self._get_temp_dir("convert_to_mac")
        shutil.rmtree(dir, ignore_errors=True)

        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        result = runner.invoke(
            convert_to_mac,
            obj={"config": ctx},
            args=["--namespace", "test-namespace", "--project-dir", dir, "--dry-run"],
        )
        self.assertEqual(result.output, f"Wrote monitor config to {dir}.\n")

        getcwd.return_value = dir
        service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            pycarlo_client=self._pycarlo_client,
            print_func=self._print_func,
        )

        files, template, _ = service.compile("test")

        self.assertEqual(len(files), 1)
        self.assertEqual(json.loads(json.dumps(template)), {"test-namespace": {"foo": "bar"}})

    @patch("os.getcwd")
    @responses.activate
    def test_convert_to_mac_with_monitors_file(self, getcwd):
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "convertUiMonitorsToConfigTemplate": {
                        "response": {
                            "configTemplateAsDict": json.dumps({"foo": "bar"}),
                            "errors": [],
                        },
                    },
                }
            },
        )

        dir = self._get_temp_dir("convert_to_mac")
        monitors_file = os.path.join(self._get_project_dir("convert_to_mac"), "monitors_uuids")
        shutil.rmtree(dir, ignore_errors=True)

        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        result = runner.invoke(
            convert_to_mac,
            obj={"config": ctx},
            args=[
                "--namespace",
                "test-namespace",
                "--project-dir",
                dir,
                "--dry-run",
                "--monitors-file",
                monitors_file,
            ],
        )
        self.assertEqual(result.output, f"Wrote monitor config to {dir}.\n")

        getcwd.return_value = dir
        service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            pycarlo_client=self._pycarlo_client,
            print_func=self._print_func,
        )

        files, template, _ = service.compile("test")

        self.assertEqual(len(files), 1)
        self.assertEqual(json.loads(json.dumps(template)), {"test-namespace": {"foo": "bar"}})

        self.assertEqual(len(responses.calls), 1)
        self.assertEqual(
            json.loads(responses.calls[0].request.body)["variables"],
            {
                "allMonitors": False,
                "dryRun": True,
                "monitorUuids": [
                    "2eb76621-3344-421a-bf8c-835173ef2d70",
                    "807cb339-1d9c-4234-a945-1e85de54513b",
                ],
                "namespace": "test-namespace",
            },
        )

    @patch("os.getcwd")
    @responses.activate
    def test_export_as_latest_with_monitors_file(self, getcwd):
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "exportLatestVersionMonitorReplacementTemplates": {
                        "configTemplateAsYaml": "foo: bar",
                        "errors": [],
                    },
                }
            },
        )

        monitors_file = os.path.join(self._get_project_dir("export_as_latest"), "monitors_uuids")

        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        result = runner.invoke(
            export_as_latest,
            obj={"config": ctx},
            args=[
                "--namespace",
                "test-namespace",
                "--monitors-file",
                monitors_file,
            ],
        )
        self.assertEqual(result.output, "foo: bar\n")

        self.assertEqual(len(responses.calls), 1)
        self.assertEqual(
            json.loads(responses.calls[0].request.body)["variables"],
            {
                "namespace": "test-namespace",
                "monitorUuids": [
                    "2eb76621-3344-421a-bf8c-835173ef2d70",
                    "807cb339-1d9c-4234-a945-1e85de54513b",
                ],
            },
        )

    @patch("os.getcwd")
    @responses.activate
    def test_get_template(self, getcwd):
        responses.post(
            _SAMPLE_CONFIG.mcd_api_endpoint,
            json={
                "data": {
                    "getMonteCarloConfigTemplates": {
                        "edges": [
                            {
                                "node": {
                                    "template": json.dumps({"foo": "bar"}),
                                }
                            }
                        ],
                    },
                }
            },
        )

        dir = self._get_temp_dir("get_template")
        shutil.rmtree(dir, ignore_errors=True)

        ctx = _SAMPLE_CONFIG
        runner = CliRunner()
        result = runner.invoke(
            get_template,
            obj={"config": ctx},
            args=["--namespace", "test-namespace", "--project-dir", dir],
        )
        self.assertEqual(result.output, f"Wrote monitor config to {dir}.\n")

        getcwd.return_value = dir
        service = MonteCarloConfigService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            pycarlo_client=self._pycarlo_client,
            print_func=self._print_func,
        )

        files, template, _ = service.compile("test")

        self.assertEqual(len(files), 1)
        self.assertEqual(json.loads(json.dumps(template)), {"test-namespace": {"foo": "bar"}})

    @staticmethod
    def _mc_config_templates_response(namespace_count):
        return Box(
            {
                "get_monte_carlo_config_templates": {
                    "edges": [
                        {
                            "node": {
                                "namespace": f"namespace_{i}",
                                "last_update_time": "2000-01-01 00:00:00.000000+00:00",
                            }
                        }
                        for i in range(1, namespace_count + 1)
                    ]
                }
            }
        )

    def _get_standalone_config_compiled_template(self, namespace):
        return {
            namespace: {
                "field_health": [
                    {
                        "table": "analytics:prod.client_hub",
                        "fields": ["account_id"],
                        "yaml_line": 3,
                        "yaml_file_name": f"{self.project_dir}/dir1/dir2/monitors.yml",
                    },
                    {
                        "table": "analytics:prod.client_hub_2",
                        "fields": ["account_id"],
                        "yaml_line": 3,
                        "yaml_file_name": f"{self.project_dir}/dir1/dir2/monitors_additional.yml",
                    },
                ],
                "notifications": {
                    "email": [
                        {
                            "name": "email2",
                            "emails": ["foo2@bar.com"],
                            "audiences": ["audience2"],
                            "yaml_line": 8,
                            "yaml_file_name": f"{self.project_dir}/dir1/dir2/monitors.yml",
                        },
                        {
                            "name": "email3",
                            "emails": ["foo3@bar.com"],
                            "audiences": ["audience3"],
                            "yaml_line": 11,
                            "yaml_file_name": f"{self.project_dir}/dir1/dir2/monitors_additional.yml",
                        },
                        {
                            "name": "email1",
                            "emails": ["foo@bar.com"],
                            "audiences": ["audience1"],
                            "yaml_line": 7,
                            "yaml_file_name": f"{self.project_dir}/dir1/monitors.yml",
                        },
                    ],
                    "yaml_line": 6,
                    "yaml_file_name": f"{self.project_dir}/dir1/monitors.yml",
                    "slack": [
                        {
                            "name": "slack1",
                            "channel": "channel1",
                            "audiences": ["audience2"],
                            "yaml_line": 11,
                            "yaml_file_name": f"{self.project_dir}/dir1/monitors.yml",
                        }
                    ],
                },
                "dimension_tracking": [
                    {
                        "table": "analytics:prod.customer_360_2",
                        "field": "account_id",
                        "yaml_line": 7,
                        "yaml_file_name": f"{self.project_dir}/dir1/dir2/monitors_additional.yml",
                    },
                    {
                        "table": "analytics:prod.customer_360",
                        "field": "account_id",
                        "yaml_line": 3,
                        "yaml_file_name": f"{self.project_dir}/dir1/monitors.yml",
                    },
                ],
            },
        }



from unittest import TestCase
from unittest.mock import Mock, patch

from parameterized import parameterized

from montecarlodata.common.data import ValidationResult
from montecarlodata.common.user import UserService
from montecarlodata.integrations.onboarding.etl.dbt_cloud import (
    DbtCloudOnboardingService,
)
from montecarlodata.integrations.onboarding.fields import (
    EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
    EXPECTED_DBT_CLOUD_RESPONSE_FIELD,
)
from montecarlodata.queries.onboarding import (
    ADD_CONNECTION_MUTATION,
    TEST_DBT_CLOUD_CRED_MUTATION,
)
from montecarlodata.utils import AwsClientWrapper, GqlWrapper
from tests.test_common_user import _SAMPLE_CONFIG

_SAMPLE_WAREHOUSE = {
    "uuid": "my-snowflake-warehouse-uuid",
    "name": "snowflake",
    "connectionType": "SNOWFLAKE",
    "connections": [
        {
            "uuid": "my-snowflake-connection-uuid",
            "type": "SNOWFLAKE",
            "createdOn": "2020-03-28T23:51:04.775647+00:00",
            "jobTypes": ["metadata", "query_logs", "sql_query", "json_schema"],
            "connectionIdentifier": {"key": "account", "value": "hda34492.us-east-1"},
        }
    ],
}


class DbtCloudOnboardingTest(TestCase):
    def setUp(self) -> None:
        self._user_service_mock = Mock(autospec=UserService)
        self._request_wrapper_mock = Mock(autospec=GqlWrapper)
        self._aws_wrapper_mock = Mock(autospec=AwsClientWrapper)

        self._service = DbtCloudOnboardingService(
            _SAMPLE_CONFIG,
            command_name="test",
            request_wrapper=self._request_wrapper_mock,
            aws_wrapper=self._aws_wrapper_mock,
            user_service=self._user_service_mock,
        )
        self._service._disable_handle_errors = True

    @parameterized.expand(
        [
            ("hmac_secret_123_xyz", "wsu_12345abcde", "dbt-cloud-webhook"),
        ]
    )
    @patch.object(DbtCloudOnboardingService, "_validate_connection")
    @patch("montecarlodata.integrations.onboarding.base.prompt_connection")
    def test_onboard_dbt_cloud_webhook_connection(
        self,
        hmac_secret,
        webhook_id,
        expected_connection_type,
        prompt_mock,
        validation_mock,
    ):
        self._user_service_mock.warehouses = [_SAMPLE_WAREHOUSE]
        validation_mock.return_value = ValidationResult(has_warnings=False, credentials_key="path")

        self._service.onboard_dbt_cloud(
            warehouseName=_SAMPLE_WAREHOUSE["name"],
            webhook_hmac_secret=hmac_secret,
            webhook_id=webhook_id,
        )

        validation_mock.assert_called_once_with(
            query=TEST_DBT_CLOUD_CRED_MUTATION,
            response_field=EXPECTED_DBT_CLOUD_RESPONSE_FIELD,
            dbt_cloud_webhook_hmac_secret=hmac_secret,
            dbt_cloud_webhook_id=webhook_id,
        )

        self._request_wrapper_mock.make_request_v2.assert_called_with(
            operation=EXPECTED_ADD_CONNECTION_RESPONSE_FIELD,
            query=ADD_CONNECTION_MUTATION,
            service="onboarding_service",
            variables={
                "key": "path",
                "connectionType": expected_connection_type,
                "name": _SAMPLE_WAREHOUSE["name"],
                "dwId": "my-snowflake-warehouse-uuid",  # check that dwId is passed as a parameter
            },
        )
